#
# Copyright (c), 2016-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
import json
import dataclasses as dc
from io import IOBase, TextIOBase
from collections.abc import Iterator
from functools import partial
from typing import Any, BinaryIO, IO, Optional, TextIO, Union
from xml.etree import ElementTree

from xmlschema.exceptions import XMLSchemaTypeError, XMLSchemaValueError, XMLResourceError
from xmlschema.names import XSD_NAMESPACE, XSI_TYPE, XSD_SCHEMA
from xmlschema.aliases import ElementType, NsmapType, LocationsType, SourceArgType, \
    DecodeType, EncodeType, JsonDecodeType, XMLSourceType, SchemaType
from xmlschema.translation import gettext as _
from xmlschema.arguments import Argument, validate_type, BooleanOption, ValidationOption
from xmlschema.utils.etree import is_etree_document, etree_tostring
from xmlschema.utils.qnames import get_extended_qname, update_namespaces, get_namespace_map
from xmlschema.resources import fetch_schema_locations, XMLResource
from xmlschema.converters import ConverterType
from xmlschema.validators import XMLSchema10, XMLSchemaBase, XMLSchemaValidationError
from xmlschema.arguments import LocationsOption
from xmlschema.settings import ResourceSettings, SchemaSettings

__all__ = ('from_json', 'is_valid', 'iter_errors', 'iter_decode', 'to_dict',
           'to_etree', 'to_json', 'validate', 'XmlDocument')

RESOURCE_KWARGS = frozenset(fld.name for fld in dc.fields(ResourceSettings))
SCHEMA_KWARGS = frozenset(fld.name for fld in dc.fields(SchemaSettings))


class SchemaArgument(Argument[SchemaType]):
    _validators = partial(validate_type, types=XMLSchemaBase),

    def __set__(self, instance: Any, value: Any) -> None:
        setattr(instance, self._name, self.validated_value(value))


def get_context(xml_document: Union[XMLSourceType, XMLResource],
                schema: Optional[Union[XMLSchemaBase, SourceArgType]] = None,
                cls: Optional[type[XMLSchemaBase]] = None,
                **kwargs: Any) -> tuple[XMLResource, SchemaType]:
    """
    Get the XML document validation/decode context.

    :return: an XMLResource instance and a schema instance.
    """
    resource: XMLResource
    if not isinstance(xml_document, XMLResource):
        _kwargs = {k: kwargs[k] for k in kwargs if k in RESOURCE_KWARGS}
        resource = XMLResource(xml_document, **_kwargs)
    elif not isinstance(xml_document, XmlDocument):
        resource = xml_document
    else:
        return xml_document, xml_document.schema

    _kwargs = {k: kwargs[k] for k in kwargs if k in SCHEMA_KWARGS}
    return resource, get_resource_schema(resource, schema, cls, **_kwargs)


def get_resource_schema(resource: XMLResource,
                        schema: Optional[Union[XMLSchemaBase, SourceArgType]] = None,
                        cls: Optional[type[XMLSchemaBase]] = None,
                        validation: str = 'strict',
                        locations: Optional[LocationsType] = None,
                        use_location_hints: bool = True,
                        **kwargs: Any) -> SchemaType:
    if cls is None:
        cls = XMLSchema10
    elif not issubclass(cls, XMLSchemaBase):
        raise XMLSchemaTypeError(_("invalid schema class {!r}").format(cls))

    if isinstance(schema, XMLSchemaBase) and resource.namespace in schema.maps.namespaces:
        return schema

    if use_location_hints:
        try:
            schema_location, locations = fetch_schema_locations(resource, locations, **kwargs)
        except ValueError:
            pass
        else:
            kwargs['locations'] = locations
            if schema is None or isinstance(schema, XMLSchemaBase):
                return cls(schema_location, **kwargs)
            else:
                return cls(schema, **kwargs)

    if isinstance(schema, XMLSchemaBase):
        return schema  # fallback to a schema for a different namespace
    elif schema is not None:
        return cls(schema, locations=locations, **kwargs)
    elif XSD_NAMESPACE == resource.namespace:
        assert cls.meta_schema is not None
        return cls.meta_schema
    elif validation == 'skip' or XSI_TYPE in resource.root.attrib:
        return get_dummy_schema(resource.root.tag, cls)
    else:
        msg = _("cannot get a schema for XML data, provide a schema argument")
        raise XMLSchemaValueError(msg)


def get_dummy_schema(tag: str, cls: type[XMLSchemaBase]) -> XMLSchemaBase:
    if tag.startswith('{'):
        namespace, name = tag[1:].split('}')
    else:
        namespace, name = '', tag

    if namespace:
        return cls(
            '<xs:schema xmlns:xs="{}" targetNamespace="{}">\n'
            '    <xs:element name="{}"/>\n'
            '</xs:schema>'.format(XSD_NAMESPACE, namespace, name)
        )
    else:
        return cls(
            '<xs:schema xmlns:xs="{}">\n'
            '    <xs:element name="{}"/>\n'
            '</xs:schema>'.format(XSD_NAMESPACE, name)
        )


def get_lazy_json_encoder(errors: list[XMLSchemaValidationError]) -> type[json.JSONEncoder]:

    class JSONLazyEncoder(json.JSONEncoder):
        def default(self, obj: Any) -> Any:
            if isinstance(obj, Iterator):
                for result in obj:
                    if isinstance(result, XMLSchemaValidationError):
                        errors.append(result)
                    else:
                        return result
                return None
            return json.JSONEncoder.default(self, obj)

    return JSONLazyEncoder


def validate(xml_document: Union[XMLSourceType, XMLResource],
             schema: Optional[XMLSchemaBase] = None,
             cls: Optional[type[XMLSchemaBase]] = None,
             path: Optional[str] = None,
             schema_path: Optional[str] = None,
             use_defaults: bool = True,
             namespaces: Optional[NsmapType] = None,
             locations: Optional[LocationsType] = None,
             use_location_hints: bool = True,
             **kwargs: Any) -> None:
    """
    Validates an XML document against a schema instance. This function builds an
    :class:`XMLSchema` object for validating the XML document. Raises an
    :exc:`XMLSchemaValidationError` if the XML document is not validated against
    the schema.

    :param xml_document: can be an :class:`XMLResource` instance, a file-like object a path \
    to a file or a URI of a resource or an Element instance or an ElementTree instance or \
    a string containing the XML data. If the passed argument is not an :class:`XMLResource` \
    instance a new one is built using this and *defuse*, *timeout* and *lazy* arguments.
    :param schema: can be a schema instance or a file-like object or a file path or a URL \
    of a resource or a string containing the schema.
    :param cls: class to use for building the schema instance (for default \
    :class:`XMLSchema10` is used).
    :param path: is an optional XPath expression that matches the elements of the XML \
    data that have to be decoded. If not provided the XML root element is used.
    :param schema_path: an XPath expression to select the XSD element to use for decoding. \
    If not provided the *path* argument or the *source* root tag are used.
    :param use_defaults: defines when to use element and attribute defaults for filling \
    missing required values.
    :param namespaces: is an optional mapping from namespace prefix to URI.
    :param locations: additional schema location hints, used if a schema instance \
    has to be built.
    :param use_location_hints: for default, in case a schema instance has \
    to be built, uses also schema locations hints provided within XML data. \
    set this option to `False` to ignore these schema location hints.
    :param kwargs: other optional arguments for building :class:`XMLResource` or \
    :class:`XMLSchema` instances provided as keyword arguments.
    """
    kwargs.update(locations=locations, use_location_hints=use_location_hints)
    source, schema = get_context(xml_document, schema, cls, **kwargs)
    schema.validate(source, path, schema_path, use_defaults, namespaces,
                    use_location_hints=use_location_hints)


def is_valid(xml_document: Union[XMLSourceType, XMLResource],
             schema: Optional[XMLSchemaBase] = None,
             cls: Optional[type[XMLSchemaBase]] = None,
             path: Optional[str] = None,
             schema_path: Optional[str] = None,
             use_defaults: bool = True,
             namespaces: Optional[NsmapType] = None,
             locations: Optional[LocationsType] = None,
             use_location_hints: bool = True,
             **kwargs: Any) -> bool:
    """
    Like :meth:`validate` except that do not raise an exception but returns ``True`` if
    the XML document is valid, ``False`` if it's invalid.
    """
    kwargs.update(validation='lax', locations=locations, use_location_hints=use_location_hints)
    source, schema = get_context(xml_document, schema, cls, **kwargs)
    return schema.is_valid(source, path, schema_path, use_defaults, namespaces,
                           use_location_hints=use_location_hints)


def iter_errors(xml_document: Union[XMLSourceType, XMLResource],
                schema: Optional[XMLSchemaBase] = None,
                cls: Optional[type[XMLSchemaBase]] = None,
                path: Optional[str] = None,
                schema_path: Optional[str] = None,
                use_defaults: bool = True,
                namespaces: Optional[NsmapType] = None,
                locations: Optional[LocationsType] = None,
                use_location_hints: bool = True,
                **kwargs: Any) -> Iterator[XMLSchemaValidationError]:
    """
    Creates an iterator for the errors generated by the validation of an XML document.
    Takes the same arguments of the function :meth:`validate`.
    """
    kwargs.update(validation='lax', locations=locations, use_location_hints=use_location_hints)
    source, schema = get_context(xml_document, schema, cls, **kwargs)
    return schema.iter_errors(source, path, schema_path, use_defaults, namespaces,
                              use_location_hints=use_location_hints)


def iter_decode(xml_document: Union[XMLSourceType, XMLResource],
                schema: Optional[XMLSchemaBase] = None,
                cls: Optional[type[XMLSchemaBase]] = None,
                path: Optional[str] = None,
                validation: str = 'lax',
                locations: Optional[LocationsType] = None,
                use_location_hints: bool = True,
                **kwargs: Any) -> Iterator[Union[Any, XMLSchemaValidationError]]:
    """
    Creates an iterator for decoding an XML source to a data structure. For default
    the document is validated during the decoding phase and if it's invalid then one
    or more :exc:`XMLSchemaValidationError` instances are yielded before the decoded data.

    :param xml_document: can be an :class:`XMLResource` instance, a file-like object a path \
    to a file or a URI of a resource or an Element instance or an ElementTree instance or \
    a string containing the XML data. If the passed argument is not an :class:`XMLResource` \
    instance a new one is built using this and *defuse*, *timeout* and *lazy* arguments.
    :param schema: can be a schema instance or a file-like object or a file path or a URL \
    of a resource or a string containing the schema.
    :param cls: class to use for building the schema instance (for default uses \
    :class:`XMLSchema10`).
    :param path: is an optional XPath expression that matches the elements of the XML \
    data that have to be decoded. If not provided the XML root element is used.
    :param validation: defines the XSD validation mode to use for decode, can be \
    'strict', 'lax' or 'skip'.
    :param locations: additional schema location hints, in case a schema instance \
    has to be built.
    :param use_location_hints: for default, in case a schema instance has \
    to be built, uses also schema locations hints provided within XML data. \
    set this option to `False` to ignore these schema location hints.
    :param kwargs: other optional arguments of :meth:`XMLSchemaBase.iter_decode` \
    or for building :class:`XMLResource` or :class:`XMLSchema` instances provided \
    as keyword arguments.
    :raises: :exc:`XMLSchemaValidationError` if the XML document is invalid and \
    ``validation='strict'`` is provided.
    """
    kwargs.update(
        validation=validation,
        locations=locations,
        use_location_hints=use_location_hints
    )
    source, _schema = get_context(xml_document, schema, cls, **kwargs)
    yield from _schema.iter_decode(source, path=path, **kwargs)


def to_dict(xml_document: Union[XMLSourceType, XMLResource],
            schema: Optional[XMLSchemaBase] = None,
            cls: Optional[type[XMLSchemaBase]] = None,
            path: Optional[str] = None,
            validation: str = 'strict',
            locations: Optional[LocationsType] = None,
            use_location_hints: bool = True,
            **kwargs: Any) -> DecodeType[Any]:
    """
    Decodes an XML document to a Python's nested dictionary. Takes the same arguments
    of the function :meth:`iter_decode`, but *validation* mode defaults to 'strict'.

    :return: an object containing the decoded data. If ``validation='lax'`` is provided \
    validation errors are collected and returned in a tuple with the decoded data.
    :raises: :exc:`XMLSchemaValidationError` if the XML document is invalid and \
    ``validation='strict'`` is provided.
    """
    kwargs.update(
        validation=validation,
        locations=locations,
        use_location_hints=use_location_hints
    )
    source, _schema = get_context(xml_document, schema, cls, **kwargs)
    return _schema.decode(source, path=path, **kwargs)


def to_json(xml_document: Union[XMLSourceType, XMLResource],
            fp: Optional[IO[str]] = None,
            schema: Optional[XMLSchemaBase] = None,
            cls: Optional[type[XMLSchemaBase]] = None,
            path: Optional[str] = None,
            validation: str = 'strict',
            locations: Optional[LocationsType] = None,
            use_location_hints: bool = True,
            json_options: Optional[dict[str, Any]] = None,
            **kwargs: Any) -> JsonDecodeType:
    """
    Serialize an XML document to JSON. For default the XML data is validated during
    the decoding phase. Raises an :exc:`XMLSchemaValidationError` if the XML document
    is not validated against the schema.

    :param xml_document: can be an :class:`XMLResource` instance, a file-like object a path \
    to a file or a URI of a resource or an Element instance or an ElementTree instance or \
    a string containing the XML data. If the passed argument is not an :class:`XMLResource` \
    instance a new one is built using this and *defuse*, *timeout* and *lazy* arguments.
    :param fp: can be a :meth:`write()` supporting file-like object.
    :param schema: can be a schema instance or a file-like object or a file path or a URL \
    of a resource or a string containing the schema.
    :param cls: schema class to use for building the instance (for default uses \
    :class:`XMLSchema10`).
    :param path: is an optional XPath expression that matches the elements of the XML \
    data that have to be decoded. If not provided the XML root element is used.
    :param validation: defines the XSD validation mode to use for decode, can be \
    'strict', 'lax' or 'skip'.
    :param locations: additional schema location hints, in case the schema instance \
    has to be built.
    :param use_location_hints: for default, in case a schema instance has \
    to be built, uses also schema locations hints provided within XML data. \
    set this option to `False` to ignore these schema location hints.
    :param json_options: a dictionary with options for the JSON serializer.
    :param kwargs: optional arguments of :meth:`XMLSchemaBase.iter_decode` as keyword arguments \
    to variate the decoding process.
    :return: a string containing the JSON data if *fp* is `None`, otherwise doesn't \
    return anything. If ``validation='lax'`` keyword argument is provided the validation \
    errors are collected and returned, eventually coupled in a tuple with the JSON data.
    :raises: :exc:`XMLSchemaValidationError` if the object is not decodable by \
    the XSD component, or also if it's invalid when ``validation='strict'`` is provided.
    """
    kwargs.update(
        validation=validation,
        locations=locations,
        use_location_hints=use_location_hints
    )
    source, _schema = get_context(xml_document, schema, cls, **kwargs)
    if json_options is None:
        json_options = {}
    if 'decimal_type' not in kwargs:
        kwargs['decimal_type'] = float

    errors: list[XMLSchemaValidationError] = []

    if path is None and source.is_lazy() and 'cls' not in json_options:
        json_options['cls'] = get_lazy_json_encoder(errors)

    obj = _schema.decode(source, path=path, **kwargs)

    if isinstance(obj, tuple):
        errors.extend(obj[1])
        if fp is not None:
            json.dump(obj[0], fp, **json_options)
            return tuple(errors)
        else:
            result = json.dumps(obj[0], **json_options)
            return result, tuple(errors)
    elif fp is not None:
        json.dump(obj, fp, **json_options)
        return None if not errors else tuple(errors)
    else:
        result = json.dumps(obj, **json_options)
        return result if not errors else (result, tuple(errors))


def to_etree(obj: Any,
             schema: Optional[Union[XMLSchemaBase, SourceArgType]] = None,
             cls: Optional[type[XMLSchemaBase]] = None,
             path: Optional[str] = None,
             validation: str = 'strict',
             namespaces: Optional[NsmapType] = None,
             use_defaults: bool = True,
             converter: Optional[ConverterType] = None,
             unordered: bool = False,
             **kwargs: Any) -> EncodeType[ElementType]:
    """
    Encodes a data structure/object to an ElementTree's Element.

    :param obj: the Python object that has to be encoded to XML data.
    :param schema: can be a schema instance or a file-like object or a file path or a URL \
    of a resource or a string containing the schema. If not provided a dummy schema is used.
    :param cls: class to use for building the schema instance (for default uses \
    :class:`XMLSchema10`).
    :param path: is an optional XPath expression for selecting the element of the schema \
    that matches the data that has to be encoded. For default the first global element of \
    the schema is used.
    :param validation: the XSD validation mode. Can be 'strict', 'lax' or 'skip'.
    :param namespaces: is an optional mapping from namespace prefix to URI.
    :param use_defaults: whether to use default values for filling missing data.
    :param converter: an :class:`XMLSchemaConverter` subclass or instance to use for \
    the encoding.
    :param unordered: a flag for explicitly activating unordered encoding mode for \
    content model data. This mode uses content models for a reordered-by-model \
    iteration of the child elements.
    :param kwargs: other optional arguments of :meth:`XMLSchemaBase.iter_encode` and \
    options for the converter.
    :return: An element tree's Element instance. If ``validation='lax'`` keyword argument is \
    provided the validation errors are collected and returned coupled in a tuple with the \
    Element instance.
    :raises: :exc:`XMLSchemaValidationError` if the object is not encodable by the schema, \
    or also if it's invalid when ``validation='strict'`` is provided.
    """
    if cls is None:
        cls = XMLSchema10
    elif not issubclass(cls, XMLSchemaBase):
        raise XMLSchemaTypeError("invalid schema class %r" % cls)

    if schema is None:
        if not path:
            raise XMLSchemaTypeError("without schema a path is required "
                                     "for building a dummy schema")

        if namespaces is None:
            tag = get_extended_qname(path, {'xsd': XSD_NAMESPACE, 'xs': XSD_NAMESPACE})
        else:
            tag = get_extended_qname(path, namespaces)

        if not tag.startswith('{') and ':' in tag:
            raise XMLSchemaTypeError("without schema the path must be "
                                     "mappable to a local or extended name")

        if tag == XSD_SCHEMA:
            assert cls.meta_schema is not None
            _schema = cls.meta_schema
        else:
            _schema = get_dummy_schema(tag, cls)

    elif isinstance(schema, XMLSchemaBase):
        _schema = schema
    else:
        _schema = cls(schema)

    return _schema.encode(
        obj=obj,
        path=path,
        validation=validation,
        namespaces=namespaces,
        use_defaults=use_defaults,
        converter=converter,
        unordered=unordered,
        **kwargs
    )


def from_json(source: Union[str, bytes, IO[str]],
              schema: Optional[Union[XMLSchemaBase, SourceArgType]] = None,
              cls: Optional[type[XMLSchemaBase]] = None,
              path: Optional[str] = None,
              validation: str = 'strict',
              namespaces: Optional[NsmapType] = None,
              use_defaults: bool = True,
              converter: Optional[ConverterType] = None,
              unordered: bool = False,
              json_options: Optional[dict[str, Any]] = None,
              **kwargs: Any) -> EncodeType[ElementType]:
    """
    Deserialize JSON data to an XML Element.

    :param source: can be a string or a :meth:`read()` supporting file-like object \
    containing the JSON document.
    :param schema: an :class:`XMLSchema10` or an :class:`XMLSchema11` instance.
    :param cls: class to use for building the schema instance (for default uses \
    :class:`XMLSchema10`).
    :param path: is an optional XPath expression for selecting the element of the schema \
    that matches the data that has to be encoded. For default the first global element of \
    the schema is used.
    :param validation: the XSD validation mode. Can be 'strict', 'lax' or 'skip'.
    :param namespaces: is an optional mapping from namespace prefix to URI.
    :param use_defaults: whether to use default values for filling missing data.
    :param converter: an :class:`XMLSchemaConverter` subclass or instance to use for \
    the encoding.
    :param unordered: a flag for explicitly activating unordered encoding mode for \
    content model data. This mode uses content models for a reordered-by-model \
    iteration of the child elements.
    :param json_options: a dictionary with options for the JSON deserializer.
    :param kwargs: other optional arguments of :meth:`XMLSchemaBase.iter_encode` and \
    options for converter.
    :return: An element tree's Element instance. If ``validation='lax'`` keyword argument is \
    provided the validation errors are collected and returned coupled in a tuple with the \
    Element instance.
    :raises: :exc:`XMLSchemaValidationError` if the object is not encodable by the schema, \
    or also if it's invalid when ``validation='strict'`` is provided.
    """
    if json_options is None:
        json_options = {}

    if isinstance(source, (str, bytes)):
        obj = json.loads(source, **json_options)
    else:
        obj = json.load(source, **json_options)

    return to_etree(
        obj=obj,
        schema=schema,
        cls=cls,
        path=path,
        validation=validation,
        namespaces=namespaces,
        use_defaults=use_defaults,
        converter=converter,
        unordered=unordered,
        **kwargs
    )


class XmlDocument(XMLResource):
    """
    An XML document bound with its schema. If no schema is get from the provided
    context and validation argument is 'skip' the XML document is associated with
    a generic schema, otherwise a ValueError is raised.

    :param source: a string containing XML data or a file path or a URL or a \
    file like object or an ElementTree or an Element.
    :param schema: can be a :class:`xmlschema.XMLSchema` instance or a file-like \
    object or a file path or a URL of a resource or a string containing the XSD schema.
    :param cls: class to use for building the schema instance (for default \
    :class:`XMLSchema10` is used).
    :param validation: the XSD validation mode to use for validating the XML document, \
    that can be 'strict' (default), 'lax' or 'skip'.
    :param namespaces: is an optional mapping from namespace prefix to URI.
    :param locations: resource location hints, that can be a dictionary or a \
    sequence of couples (namespace URI, resource URL).
    :param use_location_hints: for default, in case a schema instance has \
    to be built, uses also schema locations hints provided within XML data. \
    set this option to `False` to ignore these schema location hints.
    :param kwargs: other optional arguments for building :class:`XMLResource` or \
    :class:`XMLSchema` instances provided as keyword arguments.
    """
    errors: Union[tuple[()], list[XMLSchemaValidationError]] = ()

    # Additional arguments
    schema: SchemaArgument = SchemaArgument()
    _schema: SchemaType
    validation: ValidationOption = ValidationOption(default='strict')
    locations: LocationsOption = LocationsOption(default=None)
    use_location_hints: BooleanOption = BooleanOption(default=True)

    def __init__(self, source: XMLSourceType,
                 schema: Optional[Union[XMLSchemaBase, SourceArgType]] = None,
                 cls: Optional[type[XMLSchemaBase]] = None,
                 validation: str = 'strict',
                 namespaces: Optional[NsmapType] = None,
                 locations: Optional[LocationsType] = None,
                 use_location_hints: bool = True,
                 **kwargs: Any) -> None:

        super().__init__(source, **{k: kwargs[k] for k in kwargs if k in RESOURCE_KWARGS})

        self.validation = validation
        self.use_location_hints = use_location_hints
        self._init_namespaces = get_namespace_map(namespaces)
        self.namespaces = super().get_namespaces(namespaces, root_only=True)
        self.schema = get_resource_schema(
            resource=self,
            schema=schema,
            cls=cls,
            validation=validation,
            locations=locations,
            use_location_hints=use_location_hints,
            **{k: kwargs[k] for k in kwargs if k in SCHEMA_KWARGS}
        )

        if validation == 'strict':
            self._schema.validate(self, namespaces=self.namespaces)
        elif validation == 'lax':
            self.errors = [e for e in self._schema.iter_errors(self, namespaces=self.namespaces)]
        elif validation != 'skip':
            raise XMLSchemaValueError("%r is not a validation mode" % validation)

    def get_arguments(self) -> dict[str, Any]:
        """Returns keyword arguments for rebuilding the XML document."""
        kwargs = super().get_arguments()
        kwargs.update(
            validation=self.validation,
            schema=self.schema,
            namespaces=self._init_namespaces
        )
        return kwargs

    def get_namespaces(self, namespaces: Optional[NsmapType] = None,
                       root_only: bool = True, root_default: bool = False) -> dict[str, str]:
        namespaces = get_namespace_map(namespaces)
        update_namespaces(namespaces, self.namespaces.items(), root_declarations=True)
        return super().get_namespaces(namespaces, root_only, root_default)

    def getroot(self) -> ElementType:
        """Get the root element of the XML document."""
        return self.root

    def get_etree_document(self) -> Any:
        """
        The resource as ElementTree XML document. If the resource is lazy
        raises a resource error.
        """
        if is_etree_document(self._source):
            return self._source
        elif self._lazy:
            raise XMLResourceError(
                "cannot create an ElementTree instance from a lazy XML resource"
            )
        elif hasattr(self.root, 'getroottree'):
            return self.root.getroottree()
        else:
            return ElementTree.ElementTree(self.root)

    def decode(self, **kwargs: Any) -> DecodeType[Any]:
        """
        Decode the XML document to a nested Python dictionary.

        :param kwargs: options for the decode/to_dict method of the schema instance.
        """
        if 'validation' not in kwargs:
            kwargs['validation'] = self.validation
        if 'namespaces' not in kwargs:
            kwargs['namespaces'] = self.namespaces

        obj = self._schema.to_dict(self, **kwargs)
        return obj[0] if isinstance(obj, tuple) else obj

    def to_json(self, fp: Optional[IO[str]] = None,
                json_options: Optional[dict[str, Any]] = None,
                **kwargs: Any) -> JsonDecodeType:
        """
        Converts loaded XML data to a JSON string or file.

        :param fp: can be a :meth:`write()` supporting file-like object.
        :param json_options: a dictionary with options for the JSON deserializer.
        :param kwargs: options for the decode/to_dict method of the schema instance.
        """
        if json_options is None:
            json_options = {}
        path = kwargs.pop('path', None)
        if 'validation' not in kwargs:
            kwargs['validation'] = self.validation
        if 'namespaces' not in kwargs:
            kwargs['namespaces'] = self.namespaces
        if 'decimal_type' not in kwargs:
            kwargs['decimal_type'] = float

        errors: list[XMLSchemaValidationError] = []

        if path is None and self._lazy and 'cls' not in json_options:
            json_options['cls'] = get_lazy_json_encoder(errors)
            kwargs['lazy_decode'] = True

        obj = self._schema.decode(self, path=path, **kwargs)
        if isinstance(obj, tuple):
            if fp is not None:
                json.dump(obj[0], fp, **json_options)
                obj[1].extend(errors)
                return tuple(obj[1])
            else:
                result = json.dumps(obj[0], **json_options)
                obj[1].extend(errors)
                return result, tuple(obj[1])

        elif fp is not None:
            json.dump(obj, fp, **json_options)
            return None if not errors else tuple(errors)
        else:
            result = json.dumps(obj, **json_options)
            return result if not errors else (result, tuple(errors))

    def write(self, file: Union[str, TextIO, BinaryIO],
              encoding: str = 'us-ascii', xml_declaration: bool = False,
              default_namespace: Optional[str] = None, method: str = "xml") -> None:
        """Serialize an XML resource to a file. Cannot be used with lazy resources."""
        if self._lazy:
            raise XMLResourceError("cannot serialize a lazy XML resource")

        kwargs: dict[str, Any] = {
            'xml_declaration': xml_declaration,
            'encoding': encoding,
            'method': method,
        }
        if not default_namespace:
            kwargs['namespaces'] = self.namespaces
        else:
            namespaces: Optional[dict[Optional[str], str]]
            namespaces = {k: v for k, v in self.namespaces.items()}

            if hasattr(self.root, 'nsmap'):
                # noinspection PyTypeChecker
                namespaces[None] = default_namespace
            else:
                namespaces[''] = default_namespace
            kwargs['namespaces'] = namespaces

        _string = etree_tostring(self.root, **kwargs)

        match file:
            case str():
                if isinstance(_string, str):
                    with open(file, 'w', encoding='utf-8') as fp:
                        fp.write(_string)
                else:
                    with open(file, 'wb') as _fp:
                        _fp.write(_string)

            case TextIOBase():
                if isinstance(_string, bytes):
                    file.write(_string.decode('utf-8'))
                else:
                    file.write(_string)

            case IOBase():
                if isinstance(_string, str):
                    file.write(_string.encode('utf-8'))
                else:
                    file.write(_string)
            case _:
                msg = "unexpected type %r for 'file' argument"
                raise XMLSchemaTypeError(msg % type(file))



#
# Copyright (c), 2016-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
from collections.abc import Iterator
from functools import cached_property
from typing import cast, Any, Optional, Union

from elementpath.datatypes import AnyAtomicType

import xmlschema.names as nm
from xmlschema.exceptions import XMLSchemaValueError
from xmlschema.aliases import ElementType, NsmapType, SchemaType, ComponentClassType, \
    DecodeType, BaseXsdType, DecodedValueType, ExtraValidatorType, ValidationHookType
from xmlschema.translation import gettext as _
from xmlschema.utils.qnames import get_qname, local_name
from xmlschema.caching import schema_cache

from .exceptions import XMLSchemaCircularityError, XMLSchemaDecodeError
from .validation import ValidationContext, EncodeContext, ValidationMixin
from .helpers import parse_xsd_derivation
from .xsdbase import XSD_TYPE_DERIVATIONS, XsdComponent, XsdType
from .attributes import XsdAttributeGroup
from .assertions import XsdAssert
from .simple_types import FacetsValueType, XsdSimpleType, XsdUnion, XsdAtomic
from .groups import XsdGroup
from .wildcards import XsdAnyElement, XsdOpenContent, XsdDefaultOpenContent


class XsdComplexType(XsdType, ValidationMixin[Union[ElementType, str, bytes], Any]):
    """
    Class for XSD 1.0 *complexType* definitions.

    :var attributes: the attribute group related with the complexType.
    :var content: the content of the complexType can be a model group or a simple type.
    :var mixed: if `True` the complex type has mixed content.

    ..  <complexType
          abstract = boolean : false
          block = (#all | List of (extension | restriction))
          final = (#all | List of (extension | restriction))
          id = ID
          mixed = boolean : false
          name = NCName
          {any attributes with non-schema namespace . . .}>
          Content: (annotation?, (simpleContent | complexContent |
          ((group | all | choice | sequence)?, ((attribute | attributeGroup)*, anyAttribute?))))
        </complexType>
    """
    attributes: XsdAttributeGroup
    redefine: Optional[BaseXsdType]
    content: Union[XsdGroup, XsdSimpleType]

    abstract: bool = False
    mixed: bool = False
    assertions: Union[tuple[()], list[XsdAssert]] = ()
    open_content: Optional[XsdOpenContent] = None
    _block: Optional[str] = None

    _ADMITTED_TAGS = (nm.XSD_COMPLEX_TYPE, nm.XSD_RESTRICTION)
    _CONTENT_TAIL_TAGS = frozenset(
        (nm.XSD_ATTRIBUTE, nm.XSD_ATTRIBUTE_GROUP, nm.XSD_ANY_ATTRIBUTE)
    )

    @staticmethod
    def normalize(text: Union[str, bytes]) -> str:
        return text.decode('utf-8') if isinstance(text, bytes) else text

    __slots__ = ('content', 'attributes')

    def __init__(self, elem: ElementType,
                 schema: SchemaType,
                 parent: Optional[XsdComponent] = None,
                 name: Optional[str] = None,
                 **kwargs: Any) -> None:

        if kwargs:
            if 'content' in kwargs:
                self.content = kwargs['content']
            if 'attributes' in kwargs:
                self.attributes = kwargs['attributes']
            if 'mixed' in kwargs:
                self.mixed = kwargs['mixed']
            if 'block' in kwargs:
                self._block = kwargs['block']
            if 'final' in kwargs:
                self._final = kwargs['final']
        super().__init__(elem, schema, parent, name)

    def __repr__(self) -> str:
        if self.name is not None:
            return '%s(name=%r)' % (self.__class__.__name__, self.prefixed_name)
        elif not hasattr(self, 'content') or not hasattr(self, 'attributes'):
            return '%s(id=%r)' % (self.__class__.__name__, id(self))
        else:
            return '%s(content=%r, attributes=%r)' % (
                self.__class__.__name__, self.content_type_label,
                [a if a.name is None else a.prefixed_name for a in self.attributes.values()]
            )

    def _parse(self) -> None:
        if self.elem.tag == nm.XSD_RESTRICTION:
            return  # a local restriction is already parsed by the caller

        if 'abstract' in self.elem.attrib:
            if self.elem.attrib['abstract'].strip() in ('true', '1'):
                self.abstract = True

        if 'block' in self.elem.attrib:
            self._block = parse_xsd_derivation(self.elem, 'block', XSD_TYPE_DERIVATIONS, self)

        if 'final' in self.elem.attrib:
            self._final = parse_xsd_derivation(self.elem, 'final', XSD_TYPE_DERIVATIONS, self)

        if 'mixed' in self.elem.attrib:
            if self.elem.attrib['mixed'].strip() in ('true', '1'):
                self.mixed = True

        try:
            self.name = get_qname(self.target_namespace, self.elem.attrib['name'])
        except KeyError:
            self.name = None
            if self.parent is None:
                msg = _("missing attribute 'name' in a global complexType")
                self.parse_error(msg)
                self.name = 'nameless_%s' % str(id(self))
        else:
            if self.parent is not None:
                msg = _("attribute 'name' not allowed in a local complexType")
                self.parse_error(msg)
                self.name = None

        content_elem = self._parse_child_component(self.elem, strict=False)
        if content_elem is None or content_elem.tag in self._CONTENT_TAIL_TAGS:
            self.content = self.builders.create_empty_content_group(self)
            self._parse_content_tail(self.elem)
            default_open_content = self.default_open_content
            if default_open_content is not None and \
                    (self.mixed or self.content or default_open_content.applies_to_empty):
                self.open_content = default_open_content

        elif content_elem.tag in nm.MODEL_GROUP_TAGS:
            self.content = self.builders.group_class(content_elem, self.schema, self)
            default_open_content = self.default_open_content
            if default_open_content is not None and \
                    (self.mixed or self.content or default_open_content.applies_to_empty):
                self.open_content = default_open_content
            self._parse_content_tail(self.elem)

        elif content_elem.tag == nm.XSD_SIMPLE_CONTENT:
            if 'mixed' in content_elem.attrib:
                msg = _("'mixed' attribute not allowed with simpleContent")
                self.parse_error(msg, content_elem)

            derivation_elem = self._parse_derivation_elem(content_elem)
            if derivation_elem is None:
                return

            self.base_type = base_type = self._parse_base_type(derivation_elem)
            if derivation_elem.tag == nm.XSD_RESTRICTION:
                self._parse_simple_content_restriction(derivation_elem, base_type)
            else:
                self._parse_simple_content_extension(derivation_elem, base_type)

            if content_elem is not self.elem[-1]:
                k = 2 if content_elem is not self.elem[0] else 1
                msg = _("unexpected tag %r after simpleContent declaration:")
                self.parse_error(msg % self.elem[k].tag)

        elif content_elem.tag == nm.XSD_COMPLEX_CONTENT:
            #
            # complexType with complexContent restriction/extension
            if 'mixed' in content_elem.attrib:
                mixed = content_elem.attrib['mixed'] in ('true', '1')
                if mixed is not self.mixed:
                    self.mixed = mixed
                    if 'mixed' in self.elem.attrib and self.xsd_version == '1.1':
                        msg = _("value of 'mixed' attribute in complexType "
                                "and complexContent must be the same")
                        self.parse_error(msg)

            derivation_elem = self._parse_derivation_elem(content_elem)
            if derivation_elem is None:
                return

            self.base_type = self._parse_base_type(derivation_elem, complex_content=True)
            if self.base_type is self and self.redefine is not None:
                self.base_type = self.redefine
                self.open_content = None

            if derivation_elem.tag == nm.XSD_RESTRICTION:
                self._parse_complex_content_restriction(derivation_elem, self.base_type)
            else:
                self._parse_complex_content_extension(derivation_elem, self.base_type)

            if content_elem is not self.elem[-1]:
                k = 2 if content_elem is not self.elem[0] else 1
                msg = _("unexpected tag %r after complexContent declaration")
                self.parse_error(msg % self.elem[k].tag)

        elif content_elem.tag == nm.XSD_OPEN_CONTENT and self.xsd_version > '1.0':
            self.open_content = XsdOpenContent(content_elem, self.schema, self)

            if content_elem is self.elem[-1]:
                self.content = self.builders.create_empty_content_group(self)
            else:
                for index, child in enumerate(self.elem):
                    if content_elem is not child:
                        continue
                    elif self.elem[index + 1].tag in nm.MODEL_GROUP_TAGS:
                        self.content = self.builders.group_class(
                            self.elem[index + 1], self.schema, self
                        )
                    else:
                        self.content = self.builders.create_empty_content_group(self)
                    break
            self._parse_content_tail(self.elem)

        else:
            if self.schema.validation == 'skip':
                # Also generated by meta-schema validation for 'lax' and 'strict' modes
                msg = _("unexpected tag %r for complexType content")
                self.parse_error(msg % content_elem.tag)

            self.content = self.builders.create_any_content_group(self)
            self.attributes = self.builders.create_any_attribute_group(self)

        if self.redefine is None:
            if self.base_type is not None and self.base_type.name == self.name:
                msg = _("wrong definition with self-reference")
                self.parse_error(msg)
        elif self.base_type is None or self.base_type.name != self.name:
            msg = _("wrong redefinition without self-reference")
            self.parse_error(msg)

    def _parse_content_tail(self, elem: ElementType, **kwargs: Any) -> None:
        self.attributes = self.builders.attribute_group_class(
            elem, self.schema, self, **kwargs
        )

    def _parse_derivation_elem(self, elem: ElementType) -> Optional[ElementType]:
        derivation_elem = self._parse_child_component(elem)
        if derivation_elem is None or \
                derivation_elem.tag not in (nm.XSD_RESTRICTION, nm.XSD_EXTENSION):
            msg = _("restriction or extension tag expected")
            self.parse_error(msg, derivation_elem)
            self.content = self.builders.create_any_content_group(self)
            self.attributes = self.builders.create_any_attribute_group(self)
            return None

        if self.derivation is not None and self.redefine is None:
            msg = _("{!r} is expected to have a redefined/overridden component")
            raise XMLSchemaValueError(msg.format(self))
        self.derivation = local_name(derivation_elem.tag)

        if self.base_type is not None and self.derivation in self.base_type.final:
            msg = _("{0!r} derivation not allowed for {1!r}")
            self.parse_error(msg.format(self.derivation, self))
        return derivation_elem

    def _parse_base_type(self, elem: ElementType, complex_content: bool = False) \
            -> Union[XsdSimpleType, 'XsdComplexType']:
        try:
            base_qname = self.schema.resolve_qname(elem.attrib['base'])
        except (KeyError, ValueError, RuntimeError) as err:
            if 'base' not in elem.attrib:
                msg = _("'base' attribute required")
                self.parse_error(msg, elem)
            else:
                self.parse_error(err, elem)
            return self.maps.any_type

        try:
            base_type = self.maps.types[base_qname]
        except KeyError:
            msg = _("missing base type %r")
            self.parse_error(msg % base_qname, elem)
            if complex_content:
                return self.maps.any_type
            else:
                return self.maps.any_simple_type
        except XMLSchemaCircularityError as err:
            self.parse_error(err, err.elem)
            return self.maps.any_type
        else:
            if complex_content and base_type.is_simple():
                msg = _("a complexType ancestor required: {!r}")
                self.parse_error(msg.format(base_type), elem)
                return self.maps.any_type

            if base_type.final and elem.tag.rsplit('}', 1)[-1] in base_type.final:
                msg = _("derivation by %r blocked by attribute 'final' in base type")
                self.parse_error(msg % elem.tag.rsplit('}', 1)[-1])

            return base_type

    def _parse_simple_content_restriction(self, elem: ElementType, base_type: Any) -> None:
        # simpleContent restriction: the base type must be a complexType with a simple
        # content or a complex content with a mixed and emptiable content.
        if base_type.is_simple():
            msg = _("a complexType ancestor required: {!r}")
            self.parse_error(msg.format(base_type), elem)
            self.content = self.builders.create_any_content_group(self)
            self._parse_content_tail(elem)
        else:
            if base_type.is_empty():
                self.content = self.builders.atomic_restriction_class(
                    elem, self.schema, self
                )
                if not self.is_empty():
                    msg = _("a not empty simpleContent cannot restrict an empty content type")
                    self.parse_error(msg, elem)
                    self.content = self.builders.create_any_content_group(self)

            elif base_type.has_simple_content():
                self.content = self.builders.atomic_restriction_class(
                    elem, self.schema, self
                )
                if not self.content.is_derived(base_type.content, 'restriction'):
                    msg = _("content type is not a restriction of base content")
                    self.parse_error(msg, elem)

            elif base_type.mixed and base_type.is_emptiable():
                self.content = self.builders.atomic_restriction_class(
                    elem, self.schema, self
                )
            else:
                msg = _("with simpleContent cannot restrict an element-only content type")
                self.parse_error(msg, elem)
                self.content = self.builders.create_any_content_group(self)

            self._parse_content_tail(elem, derivation='restriction',
                                     base_attributes=base_type.attributes)

    def _parse_simple_content_extension(self, elem: ElementType, base_type: Any) -> None:
        # simpleContent extension: the base type must be a simpleType or a complexType
        # with simple content.
        child = self._parse_child_component(elem, strict=False)
        if child is not None and child.tag not in self._CONTENT_TAIL_TAGS:
            msg = _('unexpected tag %r')
            self.parse_error(msg % child.tag, child)

        if base_type.is_simple():
            self.content = base_type
            self._parse_content_tail(elem)
        else:
            if base_type.has_simple_content():
                self.content = base_type.content
            else:
                self.parse_error(_("base type %r has no simple content") % base_type, elem)
                self.content = self.builders.create_any_content_group(self)

            self._parse_content_tail(elem, derivation='extension',
                                     base_attributes=base_type.attributes)

    def _parse_complex_content_restriction(self, elem: ElementType, base_type: Any) -> None:
        if 'restriction' in base_type.final:
            msg = _("the base type is not derivable by restriction")
            self.parse_error(msg)
        if base_type.is_simple() or base_type.has_simple_content():
            msg = _("base %r is simple or has a simple content")
            self.parse_error(msg % base_type, elem)
            base_type = self.maps.any_type

        # complexContent restriction: the base type must be a complexType with a complex content.
        for child in elem:
            if child.tag == nm.XSD_OPEN_CONTENT and self.xsd_version > '1.0':
                self.open_content = XsdOpenContent(child, self.schema, self)
                continue
            elif child.tag in nm.MODEL_GROUP_TAGS:
                content = self.builders.group_class(child, self.schema, self)
                if not base_type.content.admits_restriction(content.model):
                    msg = _("restriction of an xs:{0} with more than "
                            "one particle with xs:{1} is forbidden")
                    self.parse_error(msg.format(base_type.content.model, content.model))
                break
        else:
            content = self.builders.create_empty_content_group(
                self, base_type.content.model
            )

        content.restriction = base_type.content

        if base_type.is_element_only() and content.mixed:
            msg = _("derived a mixed content from a base type that has element-only content")
            self.parse_error(msg, elem)
        elif base_type.is_empty() and not content.is_empty():
            msg = _("an empty content derivation from base type that has not empty content")
            self.parse_error(msg, elem)

        if self.open_content is None:
            default_open_content = self.default_open_content
            if default_open_content is not None and \
                    (self.mixed or content or default_open_content.applies_to_empty):
                self.open_content = default_open_content

        if self.open_content and content and \
                not self.open_content.is_restriction(base_type.open_content):
            msg = _("{0!r} is not a restriction of the base type {1!r}")
            self.parse_error(msg.format(self.open_content, base_type.open_content))

        self.content = content
        self._parse_content_tail(elem, derivation='restriction',
                                 base_attributes=base_type.attributes)

    def _parse_complex_content_extension(self, elem: ElementType, base_type: Any) -> None:
        if 'extension' in base_type.final:
            msg = _("the base type is not derivable by extension")
            self.parse_error(msg)

        group_elem: Optional[ElementType]
        for group_elem in elem:
            if group_elem.tag != nm.XSD_ANNOTATION and not callable(group_elem.tag):
                break
        else:
            group_elem = None

        if base_type.is_empty():
            if not base_type.mixed:
                # Empty element-only model extension: don't create a nested group.
                if group_elem is not None and group_elem.tag in nm.MODEL_GROUP_TAGS:
                    self.content = self.builders.group_class(
                        group_elem, self.schema, self
                    )
                elif base_type.is_simple() or base_type.has_simple_content():
                    self.content = self.builders.create_empty_content_group(self)
                else:
                    self.content = self.builders.create_empty_content_group(
                        parent=self, elem=base_type.content.elem
                    )
            else:
                # Empty mixed model extension
                self.content = self.builders.create_empty_content_group(self)
                self.content.append(self.builders.create_empty_content_group(self.content))

                if group_elem is not None and group_elem.tag in nm.MODEL_GROUP_TAGS:
                    group = self.builders.group_class(
                        group_elem, self.schema, self.content
                    )
                    if not self.mixed:
                        msg = _("base has a different content type (mixed=%r) "
                                "and the extension group is not empty.")
                        self.parse_error(msg % base_type.mixed, elem)
                else:
                    group = self.builders.create_empty_content_group(self)

                self.content.append(group)
                self.content.elem.append(base_type.content.elem)
                self.content.elem.append(group.elem)

        elif group_elem is not None and group_elem.tag in nm.MODEL_GROUP_TAGS:
            # Derivation from a simple content is forbidden if base type is not empty.
            if base_type.is_simple() or base_type.has_simple_content():
                msg = _("base %r is simple or has a simple content")
                self.parse_error(msg % base_type, elem)
                base_type = self.maps.any_type

            group = self.builders.group_class(group_elem, self.schema, self)

            if group.model == 'all':
                msg = _("cannot extend a complex content with xs:all")
                self.parse_error(msg)
            if base_type.content.model == 'all' and group.model == 'sequence':
                msg = _("xs:sequence cannot extend xs:all")
                self.parse_error(msg)

            content = self.builders.create_empty_content_group(self)
            content.append(base_type.content)
            content.append(group)
            content.elem.append(base_type.content.elem)
            content.elem.append(group.elem)

            if base_type.content.model == 'all' and base_type.content and group:
                msg = _("XSD 1.0 does not allow extension of a not empty 'all' model group")
                self.parse_error(msg)
            if base_type.mixed is not self.mixed:
                msg = _("base has a different content type (mixed=%r) "
                        "and the extension group is not empty")
                self.parse_error(msg % base_type.mixed, elem)
            self.content = content

        elif base_type.is_simple():
            self.content = base_type
        elif base_type.has_simple_content():
            self.content = base_type.content
        else:
            # Derived type has an empty content
            if self.mixed is not base_type.mixed:
                if self.mixed:
                    msg = _("extended type has a mixed content but the base is element-only")
                    self.parse_error(msg, elem)
                self.mixed = base_type.mixed  # not an error if mixed='false'

            self.content = self.builders.create_empty_content_group(self)
            self.content.append(base_type.content)
            self.content.elem.append(base_type.content.elem)

        self._parse_content_tail(elem, derivation='extension', base_attributes=base_type.attributes)

    @property
    def default_open_content(self) -> Optional[XsdDefaultOpenContent]:
        return None

    @property
    def block(self) -> str:
        return self.schema.block_default if self._block is None else self._block

    @property
    def simple_type(self) -> Optional[XsdSimpleType]:
        return self.content if isinstance(self.content, XsdSimpleType) else None

    @property
    def model_group(self) -> Optional[XsdGroup]:
        return self.content if isinstance(self.content, XsdGroup) else None

    @property
    def content_type_label(self) -> str:
        if self.is_empty():
            return 'empty'
        elif isinstance(self.content, XsdSimpleType):
            return 'simple'
        elif self.mixed:
            return 'mixed'
        else:
            return 'element-only'

    @cached_property
    def root_type(self) -> BaseXsdType:
        if self.attributes or self.base_type is None:
            return cast('XsdComplexType', self.maps.types[nm.XSD_ANY_TYPE])
        else:
            return self.base_type.root_type

    @cached_property
    def sequence_type(self) -> str:
        if self.is_empty():
            return 'empty-sequence()'
        elif isinstance(self.content, XsdAtomic):
            name = self.content.primitive_type.local_name
            st = 'item()' if name is None else f'xs:{name}'
        else:
            st = 'xs:untypedAtomic'

        return f"{st}{'*' if self.is_emptiable() else '+'}"

    @staticmethod
    def is_simple() -> bool:
        return False

    @staticmethod
    def is_complex() -> bool:
        return True

    def is_empty(self) -> bool:
        if self.open_content and self.open_content.mode != 'none':
            return False
        return self.content.is_empty()

    def is_emptiable(self) -> bool:
        return self.content.is_emptiable()

    def has_simple_content(self) -> bool:
        if not isinstance(self.content, XsdGroup):
            return not self.content.is_empty()
        elif self.content or self.content.mixed or self.base_type is None:
            return False
        else:
            return self.base_type.is_simple() or self.base_type.has_simple_content()

    def has_complex_content(self) -> bool:
        if not isinstance(self.content, XsdGroup):
            return False
        elif self.open_content and self.open_content.mode != 'none':
            return True
        return not self.content.is_empty()

    def has_mixed_content(self) -> bool:
        if not isinstance(self.content, XsdGroup):
            return False
        elif self.content.is_empty():
            return False
        else:
            return self.content.mixed

    def is_element_only(self) -> bool:
        if not isinstance(self.content, XsdGroup):
            return False
        elif self.content.is_empty():
            return False
        else:
            return not self.content.mixed

    def is_list(self) -> bool:
        return isinstance(self.content, XsdSimpleType) and self.content.is_list()

    def is_dynamic_consistent(self, other: Any) -> bool:
        return other.name == nm.XSD_ANY_TYPE or self.is_derived(other) or \
            isinstance(other, XsdUnion) and any(self.is_derived(mt) for mt in other.member_types)

    def validate(self, obj: Union[ElementType, str, bytes],
                 use_defaults: bool = True,
                 namespaces: Optional[NsmapType] = None,
                 max_depth: Optional[int] = None,
                 extra_validator: Optional[ExtraValidatorType] = None,
                 validation_hook: Optional[ValidationHookType] = None) -> None:
        kwargs: Any = {
            'use_defaults': use_defaults,
            'namespaces': namespaces,
            'max_depth': max_depth,
            'extra_validator': extra_validator,
            'validation_hook': validation_hook,
        }
        if not isinstance(obj, (str, bytes)):
            super().validate(obj, **kwargs)
        elif isinstance(self.content, XsdSimpleType):
            self.content.validate(obj, **kwargs)
        elif not self.mixed and self.base_type is not None:
            self.base_type.validate(obj, **kwargs)

    def is_valid(self, obj: Union[ElementType, str, bytes],
                 use_defaults: bool = True,
                 namespaces: Optional[NsmapType] = None,
                 max_depth: Optional[int] = None,
                 extra_validator: Optional[ExtraValidatorType] = None,
                 validation_hook: Optional[ValidationHookType] = None) -> bool:
        kwargs: Any = {
            'use_defaults': use_defaults,
            'namespaces': namespaces,
            'max_depth': max_depth,
            'extra_validator': extra_validator,
            'validation_hook': validation_hook,
        }
        if not isinstance(obj, (str, bytes)):
            return super().is_valid(obj, **kwargs)
        elif isinstance(self.content, XsdSimpleType):
            return self.content.is_valid(obj, **kwargs)
        else:
            return self.mixed or self.base_type is not None and \
                self.base_type.is_valid(obj, **kwargs)

    @schema_cache
    def is_derived(self, other: BaseXsdType, derivation: Optional[str] = None) -> bool:
        if derivation and derivation == self.derivation:
            derivation = None  # derivation mode checked

        if other.ref is not None:
            other = other.ref

        if self is other or self.ref is other:
            return True
        elif other.name == nm.XSD_ANY_TYPE:
            return derivation != 'extension'
        elif self.base_type is other:
            return derivation is None
        elif isinstance(other, XsdUnion):
            return any(self.is_derived(m, derivation) for m in other.member_types)
        elif self.base_type is None:
            if not self.has_simple_content():
                return False
            return isinstance(self.content, XsdSimpleType) and \
                self.content.is_derived(other, derivation)
        elif self.has_simple_content():
            return isinstance(self.content, XsdSimpleType) and \
                self.content.is_derived(other, derivation) or \
                self.base_type.is_derived(other, derivation)
        else:
            return self.base_type.is_derived(other, derivation)

    def iter_components(self, xsd_classes: ComponentClassType = None) \
            -> Iterator[XsdComponent]:
        if xsd_classes is None or isinstance(self, xsd_classes):
            yield self
        if self.attributes and self.attributes.parent is not None:
            yield from self.attributes.iter_components(xsd_classes)
        if self.content.parent is not None:
            yield from self.content.iter_components(xsd_classes)
        if self.base_type is not None and self.base_type.parent is not None:
            yield from self.base_type.iter_components(xsd_classes)

        for obj in filter(lambda x: x.base_type is self, self.assertions):
            if xsd_classes is None or isinstance(obj, xsd_classes):
                yield obj

    def get_facet(self, tag: str) -> Optional[FacetsValueType]:
        if isinstance(self.content, XsdSimpleType):
            return self.content.get_facet(tag)
        return None

    def admit_simple_restriction(self) -> bool:
        if 'restriction' in self.final:
            return False
        else:
            return self.has_simple_content() or self.mixed and self.is_emptiable()

    def has_restriction(self) -> bool:
        return self.derivation == 'restriction'

    def has_extension(self) -> bool:
        return self.derivation == 'extension'

    def text_decode(self, text: str, validation: str = 'skip',
                    context: Optional[ValidationContext] = None) -> DecodedValueType:
        if isinstance(self.content, XsdSimpleType):
            return self.content.text_decode(text, validation, context)
        else:
            return text

    def text_is_valid(self, text: str, context: Optional[ValidationContext] = None) -> bool:
        if isinstance(self.content, XsdSimpleType):
            return self.content.text_is_valid(text, context)
        elif self.mixed or not text.strip():
            return True
        else:
            return len(self.content) == 1 and isinstance(self.content[0], XsdAnyElement)

    def decode(self, obj: Union[ElementType, str, bytes], *args: Any, **kwargs: Any) \
            -> DecodeType[Any]:
        if not isinstance(obj, (str, bytes)):
            return super().decode(obj, *args, **kwargs)
        elif isinstance(self.content, XsdSimpleType):
            return self.content.decode(obj, *args, **kwargs)
        else:
            msg = _("cannot decode %(obj)r data with %(decoder)r")
            raise XMLSchemaDecodeError(
                self, obj, str, msg % {'obj': obj, 'decoder': self}
            )

    def raw_decode(self, obj: Union[ElementType, str, bytes],
                   validation: str, context: ValidationContext) -> Any:
        """
        Decodes an Element instance using a dummy XSD element. Typically used
        for decoding with xs:anyType when an XSD element is not available.
        Also decodes strings if the type has a simple content.

        :param obj: the XML data that has to be decoded.
        :param validation: the validation mode. Can be 'lax', 'strict' or 'skip.
        :param context: the decoding context.
        :return: a decoded object.
        """
        if not isinstance(obj, (str, bytes)):
            xsd_element = self.builders.create_element(
                obj.tag, self.schema, self, form='unqualified'
            )
            xsd_element.type = self
            return xsd_element.raw_decode(obj, validation, context)
        elif isinstance(self.content, XsdSimpleType):
            return self.content.raw_decode(obj, validation, context)
        else:
            msg = _("cannot decode %(obj)r data with %(decoder)r")
            raise XMLSchemaDecodeError(
                self, obj, str, msg % {'obj': obj, 'decoder': self}
            )

    def raw_encode(self, obj: Any, validation: str, context: EncodeContext) \
            -> Optional[ElementType]:
        """
        Encode XML data. A dummy element is created for the type, and it's used for
        encode data. Typically used for encoding with xs:anyType when an XSD element
        is not available.

        :param obj: decoded XML data.
        :param validation: the validation mode. Can be 'lax', 'strict' or 'skip.
        :param context: the encoding context.
        :return: returns an Element.
        """
        try:
            name, value = obj
        except ValueError:
            name = obj.name
            value = obj

        xsd_type: BaseXsdType
        if isinstance(value, AnyAtomicType):
            xsd_type = self.maps.any_atomic_type
        else:
            xsd_type = self.maps.any_type

        xsd_element = self.builders.create_element(
            name, self.schema, xsd_type, form='unqualified'
        )
        xsd_element.type = xsd_type

        return xsd_element.raw_encode(value, validation, context)


class Xsd11ComplexType(XsdComplexType):
    """
    Class for XSD 1.1 *complexType* definitions.

    ..  <complexType
          abstract = boolean : false
          block = (#all | List of (extension | restriction))
          final = (#all | List of (extension | restriction))
          id = ID
          mixed = boolean
          name = NCName
          defaultAttributesApply = boolean : true
          {any attributes with non-schema namespace . . .}>
          Content: (annotation?, (simpleContent | complexContent | (openContent?,
          (group | all | choice | sequence)?,
          ((attribute | attributeGroup)*, anyAttribute?), assert*)))
        </complexType>
    """
    default_attributes_apply = True

    _CONTENT_TAIL_TAGS = nm.CONTENT_TAIL_TAGS

    @property
    def default_attributes(self) -> Optional[XsdAttributeGroup]:
        if self.redefine is not None:
            default_attributes = self.schema.default_attributes
        else:
            for child in self.schema.root:
                if child.tag == nm.XSD_OVERRIDE and self.elem in child:
                    schema = self.schema.includes[child.attrib['schemaLocation']]
                    if schema.override is self.schema:
                        default_attributes = schema.default_attributes
                        break
            else:
                default_attributes = self.schema.default_attributes

        if isinstance(default_attributes, str):
            return None
        return default_attributes

    @property
    def default_open_content(self) -> Optional[XsdDefaultOpenContent]:
        if self.parent is not None:
            return self.schema.default_open_content

        for child in self.schema.root:
            if child.tag == nm.XSD_OVERRIDE and self.elem in child:
                schema = self.schema.includes[child.attrib['schemaLocation']]
                if schema.override is self.schema:
                    return schema.default_open_content
        else:
            return self.schema.default_open_content

    def _parse(self) -> None:
        super()._parse()

        if self.base_type and self.base_type.base_type is self.maps.any_simple_type and \
                self.base_type.derivation == 'extension' and not self.attributes:
            # Derivation from xs:anySimpleType with missing variety.
            # See: http://www.w3.org/TR/xmlschema11-1/#Simple_Type_Definition_details
            msg = _("the simple content of {!r} is not a valid simple type in XSD 1.1")
            self.parse_error(msg.format(self.base_type))

        # Add open content to a complex content type
        if isinstance(self.content, XsdGroup):
            if self.open_content is None:
                if self.content.open_content is not None:
                    msg = _("openContent mismatch between type and model group")
                    self.parse_error(msg)
            elif self.open_content:
                self.content.open_content = self.open_content

        # Add inheritable attributes
        if isinstance(self.base_type, XsdComplexType):
            for name, attr in self.base_type.attributes.items():
                if attr.inheritable:
                    if name not in self.attributes:
                        self.attributes[name] = attr
                    elif not self.attributes[name].inheritable:
                        msg = _("attribute %r must be inheritable")
                        self.parse_error(msg % name)

        if 'defaultAttributesApply' not in self.elem.attrib:
            self.default_attributes_apply = True
        elif self.elem.attrib['defaultAttributesApply'].strip() in ('false', '0'):
            self.default_attributes_apply = False
        else:
            self.default_attributes_apply = True

        # Add default attributes
        if self.default_attributes_apply and \
                isinstance(self.default_attributes, XsdAttributeGroup):
            if self.redefine is None:
                for k in self.default_attributes:
                    if k in self.attributes:
                        msg = _("default attribute {!r} is already "
                                "declared in the complex type")
                        self.parse_error(msg.format(k))

            self.attributes.update((k, v) for k, v in self.default_attributes.items())

    def _parse_complex_content_extension(self, elem: ElementType, base_type: Any) -> None:
        # Complex content extension with simple base is forbidden XSD 1.1.
        # For the detailed rule refer to XSD 1.1 documentation:
        #   https://www.w3.org/TR/2012/REC-xmlschema11-1-20120405/#sec-cos-ct-extends
        if base_type.is_simple() or base_type.has_simple_content():
            msg = _("base %r is simple or has a simple content")
            self.parse_error(msg % base_type, elem)
            base_type = self.maps.any_type

        if 'extension' in base_type.final:
            msg = _("the base type is not derivable by extension")
            self.parse_error(msg)

        # Parse openContent
        group_elem: Any
        for group_elem in elem:
            if group_elem.tag == nm.XSD_ANNOTATION or callable(group_elem.tag):
                continue
            elif group_elem.tag != nm.XSD_OPEN_CONTENT:
                break
            self.open_content = XsdOpenContent(group_elem, self.schema, self)
            try:
                any_element = base_type.open_content.any_element
                self.open_content.any_element.union(any_element)
            except AttributeError:
                pass
        else:
            group_elem = None

        if not base_type.content:
            if not base_type.mixed:
                # Empty element-only model extension: don't create a nested sequence group.
                if group_elem is not None and group_elem.tag in nm.MODEL_GROUP_TAGS:
                    self.content = self.builders.group_class(
                        group_elem, self.schema, self
                    )
                else:
                    max_occurs = base_type.content.max_occurs
                    self.content = self.builders.create_empty_content_group(
                        parent=self,
                        model=base_type.content.model,
                        minOccurs=str(base_type.content.min_occurs),
                        maxOccurs='unbounded' if max_occurs is None else str(max_occurs),
                    )

            else:
                # Empty mixed model extension
                self.content = self.builders.create_empty_content_group(self)
                self.content.append(self.builders.create_empty_content_group(self.content))

                if group_elem is not None and group_elem.tag in nm.MODEL_GROUP_TAGS:
                    group = self.builders.group_class(
                        group_elem, self.schema, self.content
                    )
                    if not self.mixed:
                        msg = _("base has a different content type (mixed=%r) "
                                "and the extension group is not empty.")
                        self.parse_error(msg % base_type.mixed, elem)
                    if group.model == 'all':
                        msg = _("cannot extend an empty mixed content with an xs:all")
                        self.parse_error(msg)
                else:
                    group = self.builders.create_empty_content_group(self)

                self.content.append(group)
                self.content.elem.append(base_type.content.elem)
                self.content.elem.append(group.elem)

        elif group_elem is not None and group_elem.tag in nm.MODEL_GROUP_TAGS:
            group = self.builders.group_class(group_elem, self.schema, self)

            if base_type.content.model != 'all':
                content = self.builders.create_empty_content_group(self)
                content.append(base_type.content)
                content.elem.append(base_type.content.elem)

                if group.model == 'all':
                    msg = _("xs:all cannot extend a not empty xs:%s")
                    self.parse_error(msg % base_type.content.model)
                else:
                    content.append(group)
                    content.elem.append(group.elem)
            else:
                content = self.builders.create_empty_content_group(
                    self, model='all', minOccurs=str(base_type.content.min_occurs)
                )
                content.extend(base_type.content)
                content.elem.extend(base_type.content.elem)

                if not group:
                    pass
                elif group.model != 'all':
                    msg = _("cannot extend a not empty 'all' model group with a different model")
                    self.parse_error(msg)
                elif base_type.content.min_occurs != group.min_occurs:
                    msg = _("when extend an xs:all group minOccurs must be the same")
                    self.parse_error(msg)
                elif base_type.mixed and not base_type.content:
                    msg = _("cannot extend an xs:all group with mixed empty content")
                    self.parse_error(msg)
                else:
                    content.extend(group)
                    content.elem.extend(group.elem)

            if base_type.mixed is not self.mixed:
                msg = _("base has a different content type (mixed=%r) "
                        "and the extension group is not empty.")
                self.parse_error(msg % base_type.mixed, elem)

            self.content = content

        elif base_type.is_simple():
            self.content = base_type
        elif base_type.has_simple_content():
            self.content = base_type.content
        else:
            # Derived type has an empty content
            if self.mixed is not base_type.mixed:
                if self.mixed:
                    msg = _("extended type has a mixed content but the base is element-only")
                    self.parse_error(msg, elem)
                self.mixed = base_type.mixed  # not an error if mixed='false'

            self.content = self.builders.create_empty_content_group(self)
            self.content.append(base_type.content)
            self.content.elem.append(base_type.content.elem)

        if self.open_content is None:
            default_open_content = self.default_open_content
            if default_open_content is not None and \
                    (self.mixed or self.content or default_open_content.applies_to_empty):
                self.open_content = default_open_content
            elif base_type.open_content is not None:
                self.open_content = base_type.open_content

        if base_type.open_content is not None and \
                self.open_content is not None and \
                self.open_content is not base_type.open_content:

            if self.open_content.mode == 'none':
                self.open_content = base_type.open_content
            elif not base_type.open_content.is_restriction(self.open_content):
                msg = _("{0!r} is not an extension of the base type {1!r}")
                self.parse_error(msg.format(self.open_content, base_type.open_content))

        self._parse_content_tail(elem, derivation='extension',
                                 base_attributes=base_type.attributes)

    def _parse_content_tail(self, elem: ElementType, **kwargs: Any) -> None:
        self.attributes = self.builders.attribute_group_class(
            elem, self.schema, self, **kwargs
        )

        self.assertions = [XsdAssert(e, self.schema, self, self)
                           for e in elem if e.tag == nm.XSD_ASSERT]
        if isinstance(self.base_type, XsdComplexType):
            self.assertions.extend(
                XsdAssert(assertion.elem, self.schema, self, self)
                for assertion in self.base_type.assertions
            )



#
# Copyright (c), 2025-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
from collections.abc import Callable, Iterator, Sequence
from functools import partial
from typing import Any, Optional
from xml.etree import ElementTree

from xmlschema.aliases import AncestorsType, IOType, IterParseType, ElementType, NsmapType
from xmlschema.exceptions import XMLResourceParseError, XMLSchemaValueError
from xmlschema.xpath import ElementPathSelector

FilterFunctionType = Callable[[ElementType, ElementType, AncestorsType], bool]
ClearFunctionType = Callable[[ElementType, ElementType, AncestorsType], None]


###
# Default filter and clear functions

def no_filter(root: ElementType, elem: ElementType,  ancestors: AncestorsType) -> bool:
    return True


def no_cleanup(root: ElementType, elem: ElementType,  ancestors: AncestorsType) -> None:
    return


def clear_elem(root: ElementType, elem: ElementType,  ancestors: AncestorsType) -> None:
    elem.clear()
    if ancestors is not None:
        if elem in ancestors[-1]:
            ancestors[-1].remove(elem)


###
# Iterparse generator functions

def generic_iterparse(fp: IOType,
                      events: Optional[Sequence[str]] = None,
                      filter_fn: Optional[FilterFunctionType] = None,
                      clear_fn: Optional[ClearFunctionType] = None,
                      ancestors: AncestorsType = None,
                      depth: int = -1,
                      limit: int = -1) -> Iterator[tuple[str, Any]]:
    """
    An event-based parser for filtering XML elements during parsing.

    :param fp: an open file-like object to read from.
    :param events: an optional sequence of events to filter on.
    :param filter_fn: a function that takes the root element, the \
    current element and an optional list of ancestors elements and \
    returns a boolean.
    :param clear_fn: a function that takes the root element, the \
    current element and an optional list of ancestors elements.
    :param ancestors: an optional sequence of ancestors to track.
    :param depth: an optional integer specifying the depth of the tree \
    at where to clean elements. The default value means no cleanup.
    :param limit: an optional integer specifying the maximum number of \
    parser events to process. The default value means no limit.
    """
    if events is None:
        events = 'start-ns', 'end-ns', 'start', 'end', 'comment', 'pi'
    elif 'start' not in events or 'end' not in events:
        events = tuple(events) + ('start', 'end')

    if filter_fn is None:
        filter_fn = no_filter
    if clear_fn is None:
        clear_fn = no_cleanup

    level = 0
    stop_node: Any = None
    root: Any = None
    node: Any

    try:
        for event, node in ElementTree.iterparse(fp, events):
            if not limit:
                raise StopIteration
            limit -= 1

            if event == 'end':
                level -= 1
                if level < depth:
                    if ancestors is not None:
                        ancestors.pop()
                elif level == depth and stop_node is node:
                    stop_node = None
                    clear_fn(root, node, ancestors)
            elif event == 'start':
                if level < depth:
                    if not level:
                        root = node
                    if ancestors is not None:
                        ancestors.append(node)
                elif level == depth and not filter_fn(root, node, ancestors):
                    stop_node = node
                    level += 1
                    continue

                level += 1
                if stop_node is None:
                    yield event, node
            else:
                yield event, node

    except SyntaxError as err:
        raise XMLResourceParseError("invalid XML syntax: {}".format(err)) from err


def iterfind_parser(path: str,
                    namespaces: Optional[NsmapType] = None,
                    ancestors: AncestorsType = None,
                    limit: int = -1) -> IterParseType:
    """
    Returns an iterparse function that yields elements that match the given path.
    """
    selector = ElementPathSelector(path, namespaces)

    def filter_fn(r: ElementType, e: ElementType, a: AncestorsType) -> bool:
        return selector.select_all or e in selector.iter_select(r)

    return partial(
        generic_iterparse,
        filter_fn=filter_fn,
        clear_fn=clear_elem,
        ancestors=ancestors,
        depth=selector.depth,
        limit=limit
    )


def limited_parser(limit: int,
                   ancestors: AncestorsType = None,
                   depth: int = -1) -> IterParseType:
    """
    Returns an iterparse function that process a limited number of parser events.
    """
    if limit < 0:
        raise XMLSchemaValueError("limit argument must be >= 0")
    clear_fn = None if depth > 0 else clear_elem

    return partial(
        generic_iterparse,
        filter_fn=None,
        clear_fn=clear_fn,
        ancestors=ancestors,
        depth=depth,
        limit=limit,
    )



#
# Copyright (c), 2016-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
from typing import Optional

from xmlschema.names import XSD_NOTATION
from xmlschema.translation import gettext as _
from xmlschema.utils.qnames import get_qname

from .xsdbase import XsdComponent


class XsdNotation(XsdComponent):
    """
    Class for XSD *notation* declarations.

    ..  <notation
          id = ID
          name = NCName
          public = token
          system = anyURI
          {any attributes with non-schema namespace}...>
          Content: (annotation?)
        </notation>
    """
    _ADMITTED_TAGS = XSD_NOTATION,

    def _parse(self) -> None:
        if self.parent is not None:
            self.parse_error(_("a notation declaration must be global"))
        try:
            self.name = get_qname(self.target_namespace, self.elem.attrib['name'])
        except KeyError:
            self.parse_error(_("a notation must have a 'name' attribute"))

        if 'public' not in self.elem.attrib and 'system' not in self.elem.attrib:
            self.parse_error(_("a notation must have a 'public' or a 'system' attribute"))

    @property
    def public(self) -> Optional[str]:
        return self.elem.get('public')

    @property
    def system(self) -> Optional[str]:
        return self.elem.get('system')



#
# Copyright (c), 2016-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
import logging
import warnings
import functools
from operator import attrgetter
from types import MappingProxyType
from typing import Any, Optional, TYPE_CHECKING, cast

from xmlschema.aliases import SchemaType, SourceArgType, LocationsType
from xmlschema.exceptions import XMLSchemaTypeError, XMLSchemaValueError, \
    XMLResourceBlocked, XMLResourceForbidden, XMLResourceError, XMLResourceParseError
from xmlschema.translation import gettext as _
from xmlschema.utils.urls import normalize_url
from xmlschema.utils.etree import iter_schema_declarations
from xmlschema.arguments import Option, validate_subclass
from xmlschema.locations import NamespaceResourcesMap, get_locations, \
    LOCATIONS, FALLBACK_LOCATIONS

import xmlschema.names as nm

from xmlschema.validators import GlobalMaps, XMLSchemaParseError, XMLSchemaIncludeWarning, \
    XMLSchemaImportWarning

if TYPE_CHECKING:
    from xmlschema.validators import XsdGlobals  # noqa:F401

logger = logging.getLogger('xmlschema')
base_url_attribute = attrgetter('name')


class SchemaLoader:
    """
    The default schema loader, that processes an import statement only
    if the referred namespace is not imported yet.
    """
    fallback_locations = MappingProxyType({**LOCATIONS, **FALLBACK_LOCATIONS})

    locations: NamespaceResourcesMap[str]
    schema_class: type[SchemaType]
    missing_locations: set[str]  # Missing or failing resource locations

    __slots__ = ('maps', 'namespaces', 'locations', 'missing_locations', '__dict__')

    def __init__(self,
                 maps: 'XsdGlobals',
                 locations: Optional[LocationsType] = None,
                 use_fallback: bool = True) -> None:

        self.maps = maps
        self.namespaces = maps.namespaces
        self.schema_class = type(maps.validator)
        self.locations = get_locations(locations, maps.validator.base_url)
        self.missing_locations = set()

        if not use_fallback:
            self.fallback_locations = MappingProxyType({})

    def clear(self) -> None:
        self.maps.clear()
        self.missing_locations.clear()

    def get_locations(self, namespace: str, location: Optional[str] = None) -> list[str]:
        locations: list[str] = [] if location is None else [location]
        if namespace in self.locations:
            locations.extend(self.locations[namespace])

        if namespace in self.fallback_locations:
            values = self.fallback_locations[namespace]
            if isinstance(values, str):
                locations.append(values)
            else:
                locations.extend(values)

        return locations

    def load_declared_schemas(self, schema: SchemaType,
                              other_sources: Optional[list[SourceArgType]] = None) -> None:
        """
        Processes xs:include, xs:redefine, xs:override and xs:import statements,
        loading the schemas and/or the namespaced referred into declarations.
        """
        if schema not in self.maps.schemas or schema.maps is not self.maps:
            raise XMLSchemaValueError(f"{schema} is not owned by {self.maps}!")

        logger.debug("Processes inclusions and imports of schema %r", self)
        schema.imported_namespaces.clear()
        base_url = schema.base_url

        for elem in iter_schema_declarations(schema.source.root):
            location = elem.get('schemaLocation')
            if elem.tag == nm.XSD_IMPORT:
                namespace = elem.get('namespace', '').strip()
                if namespace == schema.target_namespace:
                    if namespace == '':
                        msg = _("if the 'namespace' attribute is not present on "
                                "the import statement then the imported schema "
                                "must have a 'targetNamespace'")
                    else:
                        msg = _("the attribute 'namespace' must be different "
                                "from schema's 'targetNamespace'")
                    schema.parse_error(msg)
                    continue

                schema.imported_namespaces.append(namespace)
                url = normalize_url(location, base_url) if location else None
                if self.is_missing(namespace, location, base_url):
                    self.import_namespace(schema, namespace, url)

            elif location is not None:
                # Process xs:include/xs:redefine/xs:override only if it has a
                # schemaLocation attribute. If not an error has been already
                # generated by the XSD validation with against the meta-schema.

                operation = elem.tag.split('}')[-1]
                logger.info("Process xs:%s schema from %s", operation, location)
                try:
                    _schema = self.include_schema(schema, location, base_url)
                except OSError as err:
                    # It is not an error if the location fails to resolve:
                    #   https://www.w3.org/TR/2012/REC-xmlschema11-1-20120405/#compound-schema
                    #   https://www.w3.org/TR/2012/REC-xmlschema11-1-20120405/#src-include
                    self.missing_locations.add(location)
                    if elem.tag == nm.XSD_INCLUDE:
                        msg = _("Include schema failed: {}").format(err)
                    else:
                        # If the xs:redefine/xs:override doesn't contain components (annotation
                        # excluded) the statement is equivalent to an include, so no error is
                        # generated, otherwise fails.
                        if any(e.tag != nm.XSD_ANNOTATION and not callable(e.tag) for e in elem):
                            schema.parse_error(str(err), elem)

                        if elem.tag == nm.XSD_REDEFINE:
                            msg = _("Redefine schema failed: {}").format(err)
                        else:
                            msg = _("Override schema failed: {}").format(err)

                    schema.warnings.append(msg)
                    warnings.warn(msg, XMLSchemaIncludeWarning, stacklevel=3)

                except (XMLSchemaParseError, XMLSchemaTypeError, XMLResourceParseError) as err:
                    if elem.tag == nm.XSD_INCLUDE:
                        msg = _("can't include schema {!r}: {}")
                    elif elem.tag == nm.XSD_REDEFINE:
                        msg = _("can't redefine schema {!r}: {}")
                    else:
                        msg = _("can't override schema {!r}: {}")

                    if isinstance(err, (XMLSchemaParseError, XMLResourceParseError)):
                        schema.parse_error(msg.format(location, err), elem)
                    else:
                        raise type(err)(msg.format(location, err))
                else:
                    if _schema is schema and operation != 'include':
                        msg = _("can't {} the same schema {!r}").format(operation, schema)
                        schema.parse_error(msg, elem)
                    elif operation == 'redefine':
                        _schema.redefine = schema
                    elif operation == 'override':
                        _schema.override = schema

        logger.debug("Inclusions and imports of schema %r processed", schema)

        # Import other namespaces provided by argument (usually from xsi:schemaLocation attribute).
        if self.maps.validator is schema:
            for ns in self.locations:
                if ns not in self.maps.namespaces:
                    self.import_namespace(schema, ns)

        # Add explicitly provided other schemas
        if other_sources:
            for other in other_sources:
                schema.add_schema(other, base_url=base_url)

    def import_namespace(self, schema: SchemaType,
                         namespace: str,
                         location: Optional[str] = None) -> None:
        import_error: Optional[Exception] = None
        locations = self.get_locations(namespace, location)

        for url in locations:
            try:
                logger.info("Import namespace %r from %r", namespace, url)
                self.import_schema(schema, namespace, url, schema.base_url)
            except (OSError, XMLResourceBlocked, XMLResourceForbidden) as err:
                # It's not an error if the location access fails (ref. section 4.2.6.2):
                #   https://www.w3.org/TR/2012/REC-xmlschema11-1-20120405/#composition-schemaImport
                #
                # Also consider block or defuse of XML data as a location access fail.
                self.missing_locations.add(url)
                logger.debug('%s', err)
                if import_error is None:
                    import_error = err
            except (XMLSchemaParseError, XMLSchemaTypeError, XMLResourceParseError) as err:
                if namespace:
                    msg = _("import of namespace {!r} failed: {}").format(namespace, err)
                else:
                    msg = _("import of chameleon schema failed: {}").format(err)

                if isinstance(err, (XMLSchemaParseError, XMLResourceParseError)):
                    schema.parse_error(msg)
                else:
                    raise type(err)(msg)

            except XMLSchemaValueError as err:
                schema.parse_error(err)
            else:
                logger.info("Namespace %r imported from %r", namespace, url)
                break
        else:
            if import_error is not None:
                msg = _("Import of namespace {!r} from {!r} failed: {}.")
                schema.warnings.append(msg.format(namespace, locations, str(import_error)))
                warnings.warn(schema.warnings[-1], XMLSchemaImportWarning, stacklevel=4)

    def import_schema(self, target_schema: SchemaType,
                      namespace: str,
                      location: str,
                      base_url: Optional[str] = None,
                      build: bool = False,
                      partial: bool = False) -> Optional[SchemaType]:
        """
        Imports a schema from a specific URL.

        :param target_schema: the importer schema.
        :param namespace: is the URI of the external namespace.
        :param location: is the URL of the schema.
        :param base_url: is an optional base URL for fetching the schema resource, \
        otherwise use the base url of the loader, if any.
        :param build: defines when to build the imported schema, the default is to not build.
        :param partial: if `True`, the loaded schema is initialized without processing \
        imports/inclusions and the build phase is skipped.
        :return: the imported :class:`XMLSchema` instance.
        """
        if location in target_schema.imports:
            schema = target_schema.imports[location]
            if schema is not None and schema.target_namespace != namespace:
                msg = _('imported schema {!r} has an unmatched namespace {!r}')
                raise XMLSchemaValueError(msg.format(location, namespace))

        if namespace in self.namespaces:
            logger.debug("Import schema in namespace %r from %r", namespace, location)
        else:
            logger.debug("Import namespace %r from %r", namespace, location)

        schema = self.load_schema(location, namespace, base_url, build, partial)
        if target_schema is schema:
            return schema
        elif location not in target_schema.imports:
            target_schema.imports[location] = schema

        return schema

    def include_schema(self, target_schema: SchemaType,
                       location: str,
                       base_url: Optional[str] = None,
                       build: bool = False,
                       partial: bool = False) -> SchemaType:
        """
        Includes a schema for the same namespace from a specific location.

        :param target_schema: the schema that includes the schema at provided location.
        :param location: is the URL of the schema.
        :param base_url: is an optional base URL for fetching the schema resource.
        :param build: defines when to build the imported schema, the default is to not build.
        :return: the included :class:`XMLSchema` instance.
        :param partial: if `True`, the included schema is initialized without processing \
        imports/inclusions and the build phase is skipped.
        :return: the included :class:`XMLSchema` instance.
        """
        if location in target_schema.includes:
            return target_schema.includes[location]

        namespace = target_schema.target_namespace
        logger.debug("Include schema in namespace %r from %r", namespace, location)
        schema = self.load_schema(location, namespace, base_url, build, partial)

        if target_schema is not schema:
            target_schema.includes[location] = schema
        return schema

    def load_schema(self, source: SourceArgType,
                    namespace: Optional[str] = None,
                    base_url: Optional[str] = None,
                    build: bool = False,
                    partial: bool = False) -> SchemaType:
        """
        Loads a schema from a location.

        :param source: a URI that reference to a resource or a file path or a file-like \
        object or a string containing the schema or an Element or an ElementTree document.
        :param namespace: is the URI of the namespace that the schema belongs to.
        :param base_url: is an optional base URL for fetching the schema resource, \
        otherwise use the base url of the loader, if any.
        :param build: defines when to build the loaded schema, the default is to not build.
        :param partial: if `True`, the loaded schema is initialized without processing \
        imports/inclusions and the build phase is skipped.
        :return: the loaded schema or the schema that matches the URL if it's already loaded.
        """
        schema = self.maps.get_schema(namespace, source, base_url)
        if schema is not None:
            logger.info("Resource %r is already loaded", schema.source)
            return schema

        return self.schema_class(
            source=source,
            namespace=namespace,
            validation=self.maps.validation,
            global_maps=self.maps,
            base_url=base_url or self.maps.settings.base_url,
            build=build,
            partial=partial,
        )

    def load_namespace(self, namespace: str, build: bool = True) -> bool:
        """
        Load namespace from available location hints. Returns `True` if the namespace
        is already loaded or if the namespace can be loaded from one of the locations,
        returns `False` otherwise. Failing locations are inserted into the missing
        locations list.

        :param namespace: the namespace to load.
        :param build: if left with `True` value builds the maps after load. If the \
        build fails the resource URL is added to missing locations and the global \
        maps are restored at previous state.
        """
        if namespace in self.maps.namespaces:
            return True
        elif self.maps.validator.meta_schema is None:
            return True  # Do not load additional namespaces in meta-schema maps

        if not build:
            for url in self.get_locations(namespace):
                if self.is_missing(namespace, url):
                    try:
                        self.load_schema(url, namespace)
                    except OSError:
                        self.missing_locations.add(url)
            return namespace in self.namespaces

        with self.maps.protect_status(reraise=False):
            for url in self.get_locations(namespace):
                if self.is_missing(namespace, url):
                    try:
                        self.load_schema(url, namespace)
                    except OSError:
                        self.missing_locations.add(url)

            self.maps.build()

        return namespace in self.namespaces

    def is_missing(self, namespace: str, location: Optional[str] = None,
                   base_url: Optional[str] = None) -> bool:
        """Returns `True` if the given location hint is missing, `False` otherwise."""
        return namespace not in self.namespaces or \
            all(s.maps is not self.maps for s in self.namespaces[namespace])


class LocationSchemaLoader(SchemaLoader):
    """
    A schema loader that processes an import statement if the
    referred location is not already loaded.
    """
    def is_missing(self, namespace: str, location: Optional[str] = None,
                   base_url: Optional[str] = None) -> bool:
        if super().is_missing(namespace):
            return True
        elif location is None:
            return False
        else:
            url = normalize_url(location, base_url)
            return all(not s.source.match_location(url) for s in self.maps.schemas)


class SafeSchemaLoader(SchemaLoader):
    """
    A schema loader that processes an import statement if the
    referred location is not already loaded and after checking
    if there aren't collisions with loaded schemas.
    """
    def __init__(self, maps: 'XsdGlobals', *args: Any, **kwargs: Any) -> None:
        super().__init__(maps, *args, **kwargs)
        self.global_maps = GlobalMaps.from_builders(maps.validator.builders)
        self.excluded_locations: set[str] = set()

    def clear(self) -> None:
        self.maps.clear()
        self.missing_locations.clear()
        self.global_maps.clear()
        self.excluded_locations.clear()

    def is_missing(self, namespace: str, location: Optional[str] = None,
                   base_url: Optional[str] = None) -> bool:
        if super().is_missing(namespace):
            return True
        elif location is None:
            return False

        url = normalize_url(location, base_url)
        if url in self.excluded_locations or \
                any(s.source.match_location(url) for s in self.maps.schemas):
            return False

        schema = self.maps.get_schema(namespace, url)
        if schema is not None and schema.maps is self.maps:
            return False

        try:
            schema = self.load_schema(location, namespace, base_url, partial=True)
        except (OSError, XMLResourceError):
            self.missing_locations.add(url)
            return False  # The resource is not accessible

        self.global_maps.clear()
        try:
            self.global_maps.load(self.namespaces[namespace])
        except XMLSchemaParseError:
            self.namespaces[namespace].pop()
            self.maps.schemas.remove(schema)
            self.excluded_locations.add(url)
            return False
        else:
            self.namespaces[namespace].pop()
            self.maps.schemas.remove(schema)
            return True


class LoaderClassOption(Option[type[SchemaLoader]]):
    _validators = functools.partial(validate_subclass, cls=SchemaLoader, none=True),

    def validated_value(self, value: Any) -> type[SchemaLoader]:
        if value is None:
            return self._default
        return super().validated_value(value)

        return cast(type[SchemaLoader], value)



#
# Copyright (c), 2016-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
import os.path
import ntpath
import platform
import posixpath
import sys
from pathlib import PurePath, PurePosixPath, PureWindowsPath
from string import ascii_letters
from urllib.parse import urlsplit, unquote, quote_from_bytes

from xmlschema.exceptions import XMLSchemaValueError

SUPPORTS_UNC_DRIVE = sys.version_info >= (3, 12, 5)


def is_unc_path(path: str) -> bool:
    """
    Returns `True` if the provided path is a UNC path, `False` otherwise.
    Based on the capabilities of `PureWindowsPath` of the Python release.
    """
    return LocationWindowsPath(path).drive.startswith('\\\\')


def is_drive_path(path: str) -> bool:
    """Returns `True` if the provided path starts with a drive (e.g. 'C:'), `False` otherwise."""
    drive = ntpath.splitdrive(path)[0]
    return len(drive) == 2 and drive[1] == ':' and drive[0] in ascii_letters


class LocationPath(PurePath):
    """
    A version of pathlib.PurePath with an enhanced URI conversion and for
    the normalization of location paths.

    A system independent path normalization without resolution is essential for
    processing resource locations, so the use or base class internals can be
    necessary for using pathlib. Despite the URL path has to be considered
    case-sensitive (ref. https://www.w3.org/TR/WD-html40-970708/htmlweb.html)
    this not always happen. On the other hand the initial source is often a
    filepath, so the better choice is to maintain location paths still related
    to the operating system.
    """
    _path_module = os.path
    __slots__ = ()

    def __new__(cls, *args: str) -> 'LocationPath':
        if cls is LocationPath:
            cls = LocationWindowsPath if os.name == 'nt' else LocationPosixPath
        return super().__new__(cls, *args)  # type: ignore[arg-type, unused-ignore]

    @classmethod
    def from_uri(cls, uri: str) -> 'LocationPath':
        """
        Parse a URI and return a LocationPath. For non-local schemes like 'http',
        'https', etc. a LocationPosixPath is returned. For Windows related file
        paths, like a path with a drive, a UNC path or a path containing a backslash,
        a LocationWindowsPath is returned.
        """
        uri = uri.strip()
        parts = urlsplit(uri)

        if (scheme := parts.scheme) == 'urn':
            raise XMLSchemaValueError(f"Can't create a {cls!r} from an URN!")
        elif scheme and scheme != 'file' and (scheme not in ascii_letters or len(scheme) != 1):
            return LocationPosixPath(unquote(parts.path))

        path = parts.path
        if parts.netloc:
            if path and path[:1] != '/':
                path = '/' + path
            if parts.netloc.startswith(('/', '\\')):
                path = parts.netloc + path
            else:
                path = f'//{parts.netloc}{path}'

        elif not parts.scheme and path.startswith('//'):
            path = '//' + path

        if parts.query:
            path = f'{path}?{parts.query}'
        if parts.fragment:
            path = f'{path}#{parts.fragment}'

        if parts.scheme in ascii_letters and len(parts.scheme) == 1:
            # uri is a Windows path with a drive, e.g. k:/Python/lib/file
            path = f'{uri[0]}:{path}'  # urlsplit() converts scheme to lowercase
            return LocationWindowsPath(unquote(path))

        # Detect invalid Windows paths (rooted or UNC path followed by a drive)
        for k in range(len(path)):
            if path[k] not in '/\\':
                if not k or not is_drive_path(path[k:]):
                    break
                elif k == 1 and parts.scheme == 'file':
                    # Valid case for a URL with a file scheme
                    return LocationWindowsPath(unquote(path[1:]))
                else:
                    raise XMLSchemaValueError(f"Invalid URI {uri!r}")

        if '\\' in path or platform.system() == 'Windows':
            return LocationWindowsPath(unquote(path))
        elif ntpath.splitdrive(path)[0]:
            location_path = LocationWindowsPath(unquote(path))
            if location_path.drive:
                # PureWindowsPath not detects a drive in Python 3.11.x also
                # if it's detected by ntpath.splitdrive().
                return location_path

        return LocationPosixPath(unquote(path))

    def as_uri(self) -> str:
        # Implementation that maps relative paths to not RFC 8089 compliant relative
        # file URIs because urlopen() doesn't accept simple paths. For UNC paths uses
        # the format with four slashes to let urlopen() works.

        drive = self.drive
        if len(drive) == 2 and drive[1] == ':' and drive[0] in ascii_letters:
            # A Windows path with a drive: 'c:\dir\file' => 'file:///c:/dir/file'
            prefix = 'file:///' + drive
            path = self.as_posix()[2:]
        elif drive:
            # UNC format case: '\\host\dir\file' => 'file:////host/dir/file'
            prefix = 'file://'
            path = self.as_posix()
        else:
            path = self.as_posix()
            if path.startswith('/'):
                # A Windows relative path or an absolute posix path:
                #  ('\dir\file' | '/dir/file') => 'file://dir/file'
                prefix = 'file://'
            else:
                # A relative posix path: 'dir/file' => 'file:dir/file'
                prefix = 'file:'

        return prefix + quote_from_bytes(os.fsencode(path))

    def normalize(self) -> 'LocationPath':
        normalized_path = self._path_module.normpath(str(self))
        return self.__class__(normalized_path)


class LocationPosixPath(LocationPath, PurePosixPath):
    _path_module = posixpath
    __slots__ = ()


class LocationWindowsPath(LocationPath, PureWindowsPath):
    _path_module = ntpath
    __slots__ = ()



#
# Copyright (c), 2016-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
from collections.abc import MutableMapping, MutableSequence
from typing import TYPE_CHECKING, Any

from xmlschema.exceptions import XMLSchemaValueError
from xmlschema.aliases import NsmapType, BaseXsdType
from xmlschema.utils.qnames import local_name
from xmlschema.resources import XMLResource

from .base import ElementData, XMLSchemaConverter

if TYPE_CHECKING:
    from xmlschema.validators import XsdElement


class AbderaConverter(XMLSchemaConverter):
    """
    XML Schema based converter class for Abdera convention.

    ref: https://wiki.open311.org/JSON_and_XML_Conversion/#the-abdera-convention
    ref: https://cwiki.apache.org/confluence/display/ABDERA/JSON+Serialization

    :param namespaces: Map from namespace prefixes to URI.
    :param dict_class: dictionary class to use for decoded data. Default is `dict`.
    :param list_class: list class to use for decoded data. Default is `list`.
    """
    __slots__ = ()

    def __init__(self, namespaces: NsmapType | None = None,
                 dict_class: type[dict[str, Any]] | None = None,
                 list_class: type[list[Any]] | None = None,
                 **kwargs: Any) -> None:
        kwargs.update(attr_prefix='', text_key='', cdata_prefix=None)
        super().__init__(namespaces, dict_class, list_class, **kwargs)

    @property
    def xmlns_processing_default(self) -> str:
        return 'stacked' if isinstance(self.source, XMLResource) else 'none'

    @property
    def lossy(self) -> bool:
        return True  # Loss cdata parts

    @property
    def loss_xmlns(self) -> bool:
        return True

    def element_decode(self, data: ElementData, xsd_element: 'XsdElement',
                       xsd_type: BaseXsdType | None = None, level: int = 0) -> Any:
        xsd_type = xsd_type or xsd_element.type
        if xsd_type.simple_type is not None:
            children = data.text
        else:
            children = self.dict_class()
            for name, value, xsd_child in self.map_content(data.content):
                if value is None:
                    value = self.list_class()

                try:
                    children[name].append(value)
                except KeyError:
                    if isinstance(value, MutableSequence) and value:
                        children[name] = self.list_class((value,))
                    else:
                        children[name] = value
                except AttributeError:
                    children[name] = self.list_class((children[name], value))
            if not children:
                children = data.text

        result: list[Any] | dict[str, Any]
        if data.attributes:
            result = self.dict_class([
                ('attributes',
                 self.dict_class((k, v) for k, v in self.map_attributes(data.attributes)))
            ])
            if children is not None and children != []:
                result['children'] = self.list_class((children,))

        elif children is not None:
            result = children
        else:
            result = self.list_class()

        if level:
            return result
        elif self.dict_class is dict:
            return {self.map_qname(data.tag): result}
        return self.dict_class(((self.map_qname(data.tag), result),))

    def element_encode(self, obj: Any, xsd_element: 'XsdElement', level: int = 0) -> ElementData:
        if not isinstance(obj, MutableMapping):
            if not obj and isinstance(obj, MutableSequence):
                obj = None
            return ElementData(xsd_element.name, obj, None, {}, None)
        elif len(obj) != 1:
            tag = xsd_element.name
        else:
            key, value = next(iter(obj.items()))
            tag = self.unmap_qname(key)
            if xsd_element.is_matching(tag):
                obj = value
            elif not self.namespaces and local_name(tag) == xsd_element.local_name:
                obj = value
            else:
                tag = xsd_element.name

        attributes: dict[str, Any] = {}
        children: list[Any] | MutableMapping[str, Any]

        try:
            attributes.update((self.unmap_qname(k, xsd_element.attributes), v)
                              for k, v in obj['attributes'].items())
        except KeyError:
            children = obj
        else:
            children = obj.get('children', [])

        if isinstance(children, MutableMapping):
            children = [children]
        elif children and not isinstance(children[0], MutableMapping):
            if len(children) > 1:
                raise XMLSchemaValueError("Element %r should have only one child" % tag)
            else:
                return ElementData(tag, children[0], None, attributes, None)

        content = []
        for child in children:
            for name, value in child.items():
                if not isinstance(value, MutableSequence) or not value:
                    content.append((self.unmap_qname(name), value))
                elif isinstance(value[0], (MutableMapping, MutableSequence)):
                    ns_name = self.unmap_qname(name)
                    for item in value:
                        content.append((ns_name, item))
                else:
                    ns_name = self.unmap_qname(name)
                    xsd_child = xsd_element.match_child(ns_name)
                    if xsd_child is not None:
                        if xsd_child.type and xsd_child.type.is_list():
                            content.append((ns_name, value))
                        else:
                            content.extend((ns_name, item) for item in value)
                    else:
                        content.extend((ns_name, item) for item in value)

        return ElementData(tag, None, content, attributes, None)



#
# Copyright (c), 2024-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
import io
from xml.sax import SAXParseException
from xml.sax import expatreader  # type: ignore[attr-defined, unused-ignore]
from xml.dom import pulldom
from pyexpat import XMLParserType

from xmlschema.aliases import IOType
from xmlschema.exceptions import XMLSchemaTypeError, XMLSchemaValueError, \
    XMLResourceError, XMLResourceForbidden, XMLResourceOSError
from xmlschema.utils.streams import DefusableReader


class SafeExpatParser(expatreader.ExpatParser):  # type: ignore[misc, unused-ignore]
    _parser: XMLParserType

    def forbid_entity_declaration(self, name, is_parameter_entity,  # type: ignore
                                  value, base, sysid, pubid, notation_name):
        raise XMLResourceForbidden(f"Entities are forbidden (entity_name={name!r})")

    def forbid_unparsed_entity_declaration(self, name, base,  # type: ignore
                                           sysid, pubid, notation_name):
        raise XMLResourceForbidden(f"Unparsed entities are forbidden (entity_name={name!r})")

    def forbid_external_entity_reference(self, context, base, sysid, pubid):  # type: ignore
        raise XMLResourceForbidden(
            f"External references are forbidden (system_id={sysid!r}, public_id={pubid!r})"
        )  # pragma: no cover

    def reset(self) -> None:
        super().reset()
        self._parser.EntityDeclHandler = self.forbid_entity_declaration
        self._parser.UnparsedEntityDeclHandler = self.forbid_unparsed_entity_declaration
        self._parser.ExternalEntityRefHandler = self.forbid_external_entity_reference


def defuse_xml(fp: IOType, rewind: bool = True) -> IOType:
    """
    Defuses an XML source using a file-like object. For default the file-like object
    must be seekable because the file-like object is rewound to start position after
    the check. If it's not seekable, the file-like object is wrapped in a buffered
    reader if it's a `io.RawIOBase` or a `io.BufferedIOBase` object.

    :param fp: the file-like object to defuse.
    :param rewind: if `True` the file-like object is rewound after defusing.
    :return: the file-like object or its wrapper buffered reader.
    """
    if rewind and not fp.seekable():
        if isinstance(fp, io.RawIOBase):
            # Wrap a not seekable raw IO object in a BufferedReader
            fp = io.BufferedReader(fp)
        elif isinstance(fp, io.BufferedIOBase):
            # Other not seekable BufferedIOBase resources are wrapped in
            # a custom reader with an initial buffer of 64KiB bytes.
            try:
                fp = DefusableReader(fp)
            except (OSError, TypeError, ValueError) as err:
                if isinstance(err, OSError):
                    raise XMLResourceOSError(err)
                elif isinstance(err, TypeError):
                    raise XMLSchemaTypeError(err)
                else:
                    raise XMLSchemaValueError(err)
        else:
            msg = f"can't defuse {fp!r}: it can't be rewound after the check"
            raise XMLResourceError(msg)

    parser = SafeExpatParser()
    try:
        for event, node in pulldom.parse(fp, parser):
            if event == pulldom.START_ELEMENT:
                break
    except SAXParseException:
        pass  # the purpose is to defuse not to check xml source syntax
    except OSError as err:
        raise XMLResourceOSError(err)

    if rewind:
        try:
            fp.seek(0)
        except OSError as err:
            raise XMLResourceOSError(err)

    return fp



#
# Copyright (c), 2016-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
import copy
import importlib
import threading
import warnings
from collections.abc import Collection, Iterator, Iterable
from contextlib import contextmanager
from functools import cached_property
from itertools import dropwhile
from typing import Any, cast, Optional
from elementpath import XPathToken, XPath2Parser

import xmlschema.names as nm
from xmlschema.aliases import SchemaType, BaseXsdType, SchemaGlobalType, \
    SourceArgType, NsmapType, StagedItemType, ComponentClassType
from xmlschema.exceptions import XMLSchemaAttributeError, XMLSchemaTypeError, \
    XMLSchemaValueError, XMLSchemaWarning, XMLSchemaNamespaceError, XMLSchemaException
from xmlschema.translation import gettext as _
from xmlschema.utils.misc import deprecated
from xmlschema.utils.qnames import get_extended_qname
from xmlschema.utils.urls import get_url, normalize_url
from xmlschema.locations import NamespaceResourcesMap
from xmlschema.resources import XMLResource
from xmlschema.xpath import XsdAssertionXPathParser
from xmlschema.settings import SchemaSettings

from .exceptions import XMLSchemaValidatorError, XMLSchemaModelError, \
    XMLSchemaModelDepthError, XMLSchemaParseError
from .xsdbase import XsdValidator, XsdComponent
from .models import check_model
from . import XsdAttribute, XsdSimpleType, XsdComplexType, XsdElement, \
    XsdGroup, XsdIdentity, XsdUnion, XsdAtomicRestriction, \
    XsdAtomic, XsdAtomicBuiltin, XsdNotation, XsdAttributeGroup
from .builders import GLOBAL_MAP_ATTRIBUTE, GlobalMaps, TypesMap, NotationsMap, \
    AttributesMap, AttributeGroupsMap, ElementsMap, GroupsMap
from xmlschema import _limits


# Default placeholder for deprecation of argument 'validation' in XsdGlobals
_strict = type('str', (str,), {})('strict')


class XsdGlobals(XsdValidator, Collection[SchemaType]):
    """
    Mediator collection class for composing XML schema instances and provides lookup maps.
    It stores the global declarations defined in the registered schemas. Register a schema
    to add its declarations to the global maps.

    :param validator: the origin schema class/instance used for creating the global maps.
    :param validation: deprecated argument for the validation mode, now it takes the \
    validation mode of the validator.
    :param parent: an optional parent schema, that is required to be built and with \
    no use of the target namespace of the validator.
    :param loader_class: an optional subclass of :class:`SchemaLoader` to use for creating \
    the loader instance.
    :param locations: schema extra location hints, that can include custom resource locations \
    or additional namespaces to import after processing schema's import statements.
    :param use_fallback: if `True` the schema processor uses the validator fallback \
    location hints to load well-known namespaces (e.g. xhtml).
    :param use_xpath3: if `True` an XSD 1.1 schema instance uses the XPath 3 processor \
    for assertions. For default a full XPath 2.0 processor is used.
    :param kwargs: other keyword arguments passed to :class:`SchemaLoader`.
    """
    _schemas: set[SchemaType]

    settings: SchemaSettings
    namespaces: NamespaceResourcesMap[SchemaType]

    types: TypesMap
    """Global types map"""

    notations: NotationsMap
    """Notations map"""

    attributes: AttributesMap
    """Global attributes map"""

    attribute_groups: AttributeGroupsMap
    """Attribute groups map"""

    elements: ElementsMap
    """Global elements map"""

    groups: GroupsMap
    """Model groups map"""

    substitution_groups: dict[str, set[XsdElement]]
    """Substitution groups map"""

    identities: dict[str, XsdIdentity]
    """Identity constraints map"""

    xpath_parser_class: type[XPath2Parser]
    assertion_parser_class: type[XsdAssertionXPathParser]

    __slots__ = ('_build_lock', '_built', '_schemas', '_parent', 'validation',
                 'errors', 'validator', 'namespaces', 'loader', 'global_maps',
                 'types', 'notations', 'attributes', 'attribute_groups',
                 'elements', 'groups', 'substitution_groups', 'identities',
                 'xpath_parser_class', 'assertion_parser_class', 'settings', 'cache')

    def __init__(self, validator: SchemaType,
                 validation: str = _strict,
                 parent: Optional[SchemaType] = None,
                 settings: Optional[SchemaSettings] = None,
                 **kwargs: Any) -> None:

        if not isinstance(validation, _strict.__class__):
            msg = "argument 'validation' is not used and will be removed in v5.0"
            warnings.warn(msg, DeprecationWarning, stacklevel=1)

        super().__init__(validator.validation)
        self._build_lock = threading.Lock()
        self._built = False
        self._schemas = set()
        self._parent = parent

        self.validator = validator
        self.namespaces = NamespaceResourcesMap()  # Registered schemas by namespace URI

        self.global_maps = GlobalMaps.from_builders(validator.builders)
        (self.types, self.notations, self.attributes,
         self.attribute_groups, self.elements, self.groups) = self.global_maps

        self.substitution_groups = {}
        self.identities = {}

        if isinstance(settings, SchemaSettings):
            self.settings = settings
        else:
            self.settings = SchemaSettings(**kwargs)

        if self.settings.use_xpath3:
            module = importlib.import_module('xmlschema.xpath.xpath3')
            self.xpath_parser_class = module.XPath3Parser
            self.assertion_parser_class = module.XsdAssertionXPath3Parser
        else:
            self.xpath_parser_class = XPath2Parser
            self.assertion_parser_class = XsdAssertionXPathParser

        for ancestor in self.iter_ancestors():
            self._schemas.update(ancestor.maps.schemas)
            self.namespaces.update(ancestor.maps.namespaces)

            ancestor.maps.build()
            self.global_maps.update(ancestor.maps.global_maps)
            self.substitution_groups.update(ancestor.maps.substitution_groups)
            self.identities.update(ancestor.maps.identities)

        self.loader = self.settings.get_loader(self)
        self.cache = self.settings.get_cache()
        self.validator.maps = self

    @property
    def schemas(self) -> set[SchemaType]:
        return self._schemas

    @property
    def owned_schemas(self) -> set[SchemaType]:
        """Returns a set of the registered schemas owned by this instance."""
        return {s for s in self._schemas if s.maps is self}

    @property
    def parent(self) -> Optional[SchemaType]:
        return self._parent

    @cached_property
    def any_type(self) -> 'XsdComplexType':
        return self.validator.create_any_type()

    @cached_property
    def any_simple_type(self) -> 'XsdSimpleType':
        """Property that references to the xs:anySimpleType instance of the global maps."""
        return cast('XsdSimpleType', self.types[nm.XSD_ANY_SIMPLE_TYPE])

    @cached_property
    def any_atomic_type(self) -> 'XsdSimpleType':
        """Property that references to the xs:anyAtomicType instance of the global maps."""
        return cast('XsdSimpleType', self.types[nm.XSD_ANY_ATOMIC_TYPE])

    def __repr__(self) -> str:
        return '%s(validator=%r)' % (self.__class__.__name__, self.validator)

    def __setattr__(self, name: str, value: Any) -> None:
        if hasattr(self, name):
            if name == '_built':
                self.__dict__.clear()
            elif name != '_parent' and name != 'settings':
                msg = _("can't change attribute {!r} of a global maps instance")
                raise XMLSchemaAttributeError(msg.format(name))
        super().__setattr__(name, value)

    def __len__(self) -> int:
        return len(self._schemas)

    def __iter__(self) -> Iterator[SchemaType]:
        yield from self._schemas

    def __contains__(self, obj: object) -> bool:
        return obj in self._schemas

    def __getstate__(self) -> dict[str, Any]:
        return {
            a: getattr(self, a) for a in self._mro_slots() if a not in ('_build_lock', 'cache')
        }

    def __setstate__(self, state: dict[str, Any]) -> None:
        for attr, value in state.items():
            object.__setattr__(self, attr, value)
        self._build_lock = threading.Lock()
        self.cache = self.settings.get_cache()

    def __copy__(self) -> 'XsdGlobals':
        other = type(self)(
            validator=copy.copy(self.validator),
            parent=self._parent,
            settings=self.settings
        )
        other.loader.__dict__.update(self.loader.__dict__)
        other.loader.locations = self.loader.locations.copy()
        other.loader.missing_locations.update(self.loader.missing_locations)

        other.validator.maps = other
        for schema in self._schemas:
            if schema.maps is self and schema is not self.validator:
                copy.copy(schema).maps = other

        other.clear()
        return other

    copy = __copy__

    def lookup(self, tag: str, qname: str) -> SchemaGlobalType:
        """
        General lookup method for XSD global components.

        :param tag: the expanded QName of the XSD the global declaration/definition \
        (e.g. '{http://www.w3.org/2001/XMLSchema}element'), that is used to select \
        the global map for lookup.
        :param qname: the expanded QName of the component to be looked-up.
        :returns: an XSD global component.
        :raises: an XMLSchemaValueError if the *tag* argument is not appropriate for a global \
        component, an XMLSchemaKeyError if the *qname* argument is not found in the global map.
        """
        try:
            global_map = GLOBAL_MAP_ATTRIBUTE[tag](self)
        except KeyError:
            msg = _("wrong tag {!r} for an XSD global definition/declaration")
            raise XMLSchemaValueError(msg.format(tag)) from None
        else:
            return cast(SchemaGlobalType, global_map[qname])

    @deprecated('5.0', alt='use item lookup on notations instead')
    def lookup_notation(self, qname: str) -> XsdNotation:
        return self.notations[qname]

    @deprecated('5.0', alt='use item lookup on types instead')
    def lookup_type(self, qname: str) -> BaseXsdType:
        return self.types[qname]

    @deprecated('5.0', alt='use item lookup on attributes instead')
    def lookup_attribute(self, qname: str) -> XsdAttribute:
        return self.attributes[qname]

    @deprecated('5.0', alt='use item lookup on attribute_groups instead')
    def lookup_attribute_group(self, qname: str) -> XsdAttributeGroup:
        return self.attribute_groups[qname]

    @deprecated('5.0', alt='use item lookup on groups instead')
    def lookup_group(self, qname: str) -> XsdGroup:
        return self.groups[qname]

    @deprecated('5.0', alt='use item lookup on elements instead')
    def lookup_element(self, qname: str) -> XsdElement:
        return self.elements[qname]

    def get_instance_type(self, type_name: str, base_type: BaseXsdType,
                          namespaces: NsmapType) -> BaseXsdType:
        """
        Returns the instance XSI type from global maps, validating it with the reference base type.

        :param type_name: the XSI type attribute value, a QName in prefixed format.
        :param base_type: the XSD from which the instance type has to be derived.
        :param namespaces: a mapping from prefixes to namespaces.
        """
        if isinstance(base_type, XsdComplexType) and nm.XSI_TYPE in base_type.attributes:
            xsd_attribute = cast(XsdAttribute, base_type.attributes[nm.XSI_TYPE])
            xsd_attribute.validate(type_name)

        extended_name = get_extended_qname(type_name, namespaces)
        xsi_type = self.types[extended_name]
        if xsi_type.is_derived(base_type):
            return xsi_type
        elif isinstance(base_type, XsdSimpleType) and \
                base_type.is_union() and not base_type.facets:
            # Can be valid only if the union doesn't have facets, see:
            #   https://www.w3.org/Bugs/Public/show_bug.cgi?id=4065
            if isinstance(base_type, XsdAtomicRestriction) and \
                    isinstance(base_type.primitive_type, XsdUnion):
                if xsi_type in base_type.primitive_type.member_types:
                    return xsi_type
            elif isinstance(base_type, XsdUnion):
                if xsi_type in base_type.member_types:
                    return xsi_type

        msg = _("{0!r} cannot substitute {1!r}")
        raise XMLSchemaTypeError(msg.format(xsi_type, base_type))

    @property
    def built(self) -> bool:
        return self._built

    @cached_property
    def validation_attempted(self) -> str:
        if not any(m for m in self.global_maps):
            return 'none'
        elif not self._built or any(m.total_staged for m in self.global_maps):
            return 'partial'
        else:
            return 'full'

    @cached_property
    def validity(self) -> str:
        if self.validation == 'skip':
            return 'notKnown'
        elif any(s.errors for s in self._schemas) or \
                any(c.errors for c in self.iter_components()):
            return 'invalid'
        elif not self._built or any(m.total_staged for m in self.global_maps):
            return 'notKnown'
        else:
            return 'valid'

    @cached_property
    def xpath_constructors(self) -> dict[str, type[XPathToken]]:
        if not self._built:
            return {}

        xpath_parser = self.xpath_parser_class()
        xpath_parser.schema = self.validator.xpath_proxy

        constructors: dict[str, type[XPathToken]] = {}
        for name, xsd_type in self.types.items():
            if isinstance(xsd_type, XsdAtomic) and \
                    not isinstance(xsd_type, XsdAtomicBuiltin) and \
                    name not in (nm.XSD_ANY_ATOMIC_TYPE, nm.XSD_NOTATION):
                constructors[name] = xpath_parser.schema_constructor(name)
        return constructors

    @property
    def use_meta(self) -> bool:
        return self.validator.meta_schema is None or \
            self.validator.meta_schema in self._schemas

    @property
    def unbuilt(self) -> list[XsdComponent]:
        """Property that returns a list with unbuilt components."""
        return [c for c in self.iter_components() if not c.built]

    @property
    def xsd_version(self) -> str:
        return self.validator.XSD_VERSION

    @property
    def all_errors(self) -> list[XMLSchemaParseError]:
        return [e for s in self._schemas for e in s.all_errors]

    @property
    def total_errors(self) -> int:
        return sum(s.total_errors for s in self._schemas)

    def create_bindings(self, *bases: type[Any], **attrs: Any) -> None:
        """Creates data object bindings for the XSD elements of built schemas."""
        for xsd_element in self.iter_components(xsd_classes=XsdElement):
            assert isinstance(xsd_element, XsdElement)
            if xsd_element.target_namespace != nm.XSD_NAMESPACE:
                xsd_element.get_binding(*bases, replace_existing=True, **attrs)

    def clear_bindings(self) -> None:
        for xsd_element in self.iter_components(xsd_classes=XsdElement):
            assert isinstance(xsd_element, XsdElement)
            xsd_element.binding = None

    def iter_components(self, xsd_classes: Optional[ComponentClassType] = None) \
            -> Iterator[XsdComponent]:
        """Creates an iterator for the XSD components of built schemas."""
        for xsd_global in self.global_maps.iter_globals():
            yield from xsd_global.iter_components(xsd_classes)

    def iter_globals(self) -> Iterator[SchemaGlobalType]:
        """Creates an iterator for the built XSD global components."""
        return self.global_maps.iter_globals()

    def iter_staged(self) -> Iterator[StagedItemType]:
        """Creates an iterator for the unbuilt XSD global components."""
        return self.global_maps.iter_staged()

    def iter_schemas(self) -> Iterator[SchemaType]:
        """Creates an iterator for the registered schemas."""
        yield from self._schemas

    def iter_ancestors(self) -> Iterator[SchemaType]:
        ancestors: list[SchemaType] = []
        parent = self._parent
        while parent is not None:
            ancestors.append(parent)
            parent = parent.maps.parent

        yield from reversed(ancestors)

    def get_schema(self, namespace: Optional[str] = None,
                   source: Optional[SourceArgType] = None,
                   base_url: Optional[str] = None) -> Optional[SchemaType]:
        schemas: Optional[list[SchemaType]]

        if namespace is not None:
            schemas = self.namespaces.get(namespace)
        elif isinstance(source, XMLResource):
            namespace = source.root.get('targetNamespace', '')
            schemas = self.namespaces.get(namespace)
        elif source is None:
            return None
        else:
            schemas = list(self._schemas)

        if schemas is not None:
            if source is None:
                return schemas[0]
            elif (url := get_url(source)) is not None:
                url = normalize_url(url, base_url)
                for schema in schemas:
                    if schema.source.match_location(url):
                        return schema

            elif isinstance(source, XMLResource):
                for schema in schemas:
                    if schema.source.source is source.source:
                        return schema
            else:
                for schema in schemas:
                    if schema.source.source is source:
                        return schema

        return None

    def register(self, schema: SchemaType) -> None:
        """Registers an XMLSchema instance."""
        if schema in self._schemas:
            return

        namespace = schema.target_namespace
        source = schema.source.url or schema.source

        ns_schemas = self.namespaces.get(namespace)
        if ns_schemas is None:
            self.namespaces[namespace] = [schema]
            self._schemas.add(schema)
        elif ns_schemas[0].maps is not self:
            self._schemas.add(schema)
            ns_schemas.append(schema)
            schema.maps = self
            self.merge(ancestor=ns_schemas[0].maps.validator)
        elif self.get_schema(namespace, source) is None:
            self._schemas.add(schema)
            ns_schemas.append(schema)
        else:
            source_ref = schema.source.url or type(schema.source)
            raise XMLSchemaValueError(
                f"another schema loaded from {source_ref!r} is already registered"
            )

        if _limits.MAX_SCHEMA_SOURCES < len(self._schemas):
            raise XMLSchemaValidatorError(
                self, f"number of schema sources loaded by {self!r} exceeded"
            )

        self._built = False

    def merge(self, ancestor: SchemaType) -> None:
        """Merge the global maps until to a specific ancestor."""
        self.clear()
        for validator in dropwhile(lambda x: x is not ancestor, self.iter_ancestors()):
            maps = validator.maps
            for schema in maps._schemas:
                if schema.maps is maps:
                    namespace = schema.target_namespace
                    self._schemas.remove(schema)
                    k = self.namespaces[namespace].index(schema)

                    schema = copy.copy(schema)
                    self._schemas.add(schema)
                    self.namespaces[namespace][k] = schema
                    schema.maps = self

        self._parent = ancestor.maps._parent

    def clear(self, remove_schemas: bool = False) -> None:
        """
        Clears the instance maps and schemas.

        :param remove_schemas: removes also the schema instances, keeping only the \
        validator that created the global maps instance and schemas and namespaces \
        inherited from ancestors.
        """
        self.global_maps.clear()
        self.substitution_groups.clear()
        self.identities.clear()

        # Clear maps cache and cached properties of schemas
        self.cache.clear()
        for schema in self._schemas:
            if schema.maps is self:
                schema.clear()

        if remove_schemas:
            self._schemas.clear()
            self.namespaces.clear()

            for ancestor in self.iter_ancestors():
                self._schemas.update(ancestor.maps._schemas)
                self.namespaces.update(ancestor.maps.namespaces)

            self.register(self.validator)
            if self.validator.includes:
                self.validator.includes.clear()

        self._built = False

    def build(self) -> None:
        """
        Build the maps of XSD global definitions/declarations. The global maps are
        updated adding and building the globals of not built registered schemas.
        """
        if self._built:
            return

        with self._build_lock:
            if self._built:
                return

            self.check_loaded_schemas()
            self.clear()

            for ancestor in self.iter_ancestors():
                ancestor.maps.build()
                self.global_maps.update(ancestor.maps.global_maps)
                self.substitution_groups.update(ancestor.maps.substitution_groups)
                self.identities.update(ancestor.maps.identities)

            # Have to respect the insertion order for redefinitions/overrides
            schemas = [s for ns_schemas in self.namespaces.values()
                       for s in ns_schemas if s.maps is self]

            self.global_maps.load(schemas)
            self.types.build_builtins(self.validator)
            self.global_maps.build(schemas)

            # Update substitutes of global elements
            for name in self.substitution_groups:
                xsd_element = self.elements[name]
                assert not isinstance(xsd_element.substitutes, tuple)
                xsd_element.substitutes.update(e.name for e in xsd_element.iter_substitutes())

            self.check(schemas)

            self._built = True
            for s in schemas:
                s.clear()

            self.check_validator()

    @contextmanager
    def protect_status(self, reraise: bool = True) -> Iterator['XsdGlobals']:
        """Context manager for set a restore point in case of error."""
        schemas = self._schemas.copy()
        namespaces = self.namespaces.copy()
        global_maps = self.global_maps.copy()
        substitution_groups = self.substitution_groups.copy()
        identities = self.identities.copy()
        built = self._built

        try:
            yield self
        except XMLSchemaException:
            self.clear()
            self._schemas.clear()
            self.namespaces.clear()
            self._schemas.update(schemas)
            self.namespaces.update(namespaces)
            self.global_maps.update(global_maps)
            self.substitution_groups.update(substitution_groups)
            self.identities.update(identities)
            self._built = built
            if reraise:
                raise

    def check_loaded_schemas(self) -> None:
        """Checks the coherence of schema registrations."""
        if self.validator not in self._schemas:
            raise XMLSchemaValueError(_('global maps main validator is not registered'))

        if self._parent is None:
            self.validator.create_meta_schema(global_maps=self)

        total = 0
        for namespace, schemas in self.namespaces.items():
            total += len(schemas)
            for s in schemas:
                if s.target_namespace != namespace:
                    msg = _('schema {!r} does not belong to namespace {!r}')
                    raise XMLSchemaNamespaceError(msg.format(s, namespace))
                elif s not in self._schemas:
                    msg = _('schema {!r} does not belong to {!r}')
                    raise XMLSchemaNamespaceError(msg.format(s, self))

        if len(self._schemas) != total:
            raise XMLSchemaValueError(
                _('registered schemas do not match namespace mapped schemas')
            )

    def check(self, schemas: Optional[Iterable[SchemaType]] = None) -> None:
        """
        Checks the components of provided schemas. Used after the build of the global maps.
        For default checks all schemas and raises an exception at first error.

        :param schemas: optional argument with the set of the schemas to check.
        :raise: XMLSchemaParseError
        """
        if schemas is None:
            schemas = {s for s in self._schemas if s.maps is self}

        # Checks substitution groups circularity
        for xsd_element in self.elements.values():
            if xsd_element.name in xsd_element.substitutes:
                msg = _("circularity found for substitution group with head element {}")
                xsd_element.parse_error(msg.format(xsd_element))

        # Check redefined global groups restrictions
        for group in self.groups.values():
            assert isinstance(group, XsdGroup), _("global group not built!")
            if group.schema not in schemas or group.redefine is None:
                continue

            while group.redefine is not None:
                if not any(isinstance(e, XsdGroup) and e.name == group.name for e in group) \
                        and not group.is_restriction(group.redefine):
                    msg = _("the redefined group is an illegal restriction")
                    group.parse_error(msg)

                group = group.redefine

        # Check complex content types models restrictions
        for xsd_global in filter(lambda x: x.schema in schemas, self.iter_globals()):
            xsd_type: Any
            for xsd_type in xsd_global.iter_components(XsdComplexType):
                if not isinstance(xsd_type.content, XsdGroup):
                    continue

                if xsd_type.derivation == 'restriction':
                    base_type = xsd_type.base_type
                    if base_type and base_type.name != nm.XSD_ANY_TYPE and base_type.is_complex():
                        if not xsd_type.content.is_restriction(base_type.content):
                            msg = _("the derived group is an illegal restriction")
                            xsd_type.parse_error(msg)

                    if base_type.is_complex() and not base_type.open_content and \
                            xsd_type.open_content and xsd_type.open_content.mode != 'none':
                        _group = xsd_type.schema.builders.create_any_content_group(
                            parent=xsd_type,
                            any_element=xsd_type.open_content.any_element
                        )
                        if not _group.is_restriction(base_type.content):
                            msg = _("restriction has an open content but base type has not")
                            _group.parse_error(msg)

                try:
                    check_model(xsd_type.content)
                except XMLSchemaModelDepthError:
                    msg = _("can't verify the content model of {!r} "
                            "due to exceeding of maximum recursion depth")
                    xsd_type.schema.warnings.append(msg.format(xsd_type))
                    warnings.warn(msg, XMLSchemaWarning, stacklevel=4)
                except XMLSchemaModelError as err:
                    if self.validation == 'strict':
                        raise
                    xsd_type.errors.append(err)
